"""Base class for buffer objects.

@todo: Implement binding as separate object to allow binding buffers to
different targets. This problem is very similar to rebinding L{Sampler}s to
different units and should probably be handled in a similar way.
@todo: Implement slicing with C{glGetBufferSubData}.
@todo: Remember that C{glBindBuffer} and C{glBindBufferRange}/C{glBindBufferBase} interfer with each other!

@author: Stephan Wenger
@date: 2012-02-29

@todo: accept another buffer in constructor, copy _id, proxy shape etc., keep reference to other buffer, disable set_data and __del__
"""

import numpy as _np

import glitter.raw as _gl
from glitter.utils import primitive_types, buffer_usages, Datatype, coerce_array, ManagedObject, BindableObject

class BaseBuffer(ManagedObject, BindableObject):
    _generate_id = _gl.glGenBuffers
    _delete_id = _gl.glDeleteBuffers
    _db = "buffers"

    drawmodes = primitive_types
    usages = buffer_usages

    def __init__(self, data=None, shape=None, dtype=None, usage=None, context=None):
        if any(x is NotImplemented for x in (self._target,)):
            raise TypeError("%s is abstract" % self.__class__.__name__)
        super(BaseBuffer, self).__init__(context=context)
        if usage is None:
            usage = BaseBuffer.usages.STATIC_DRAW
        else:
            usage = self.usages(usage)
        self.set_data(data=data, shape=shape, dtype=dtype, usage=usage)

    def set_data(self, data=None, shape=None, dtype=None, usage=None):
        if data is None:
            if shape is None or dtype is None:
                raise ValueError("must specify either data or both shape and dtype")
            self._shape = shape
            self._dtype = dtype.coerced(force_gl=True)
        else:
            data = coerce_array(data, dtype, force_gl=True)
            if shape is not None:
                data = data.reshape(shape)
            self._shape = data.shape
            self._dtype = Datatype.from_numpy(data.dtype)

        if usage is None:
            usage = self.usage

        _nbytes = _np.prod(self._shape) * self._dtype.nbytes
        _data = data.ctypes if data is not None else _gl.POINTER(_gl.GLvoid)()
        with self:
            _gl.glBufferData(self._target, _nbytes, _data, usage._value)

    def get_data(self):
        _data = _np.empty(self.shape, dtype=self.dtype.as_numpy())
        with self:
            _gl.glGetBufferSubData(self._target, 0, _data.nbytes, _data.ctypes)
        return _data

    @property
    def data(self):
        return self.get_data()

    @data.setter
    def data(self, data):
        self.set_data(data)

    @property
    def shape(self):
        return self._shape

    def __len__(self):
        return self.shape[0]

    @property
    def dtype(self):
        return self._dtype

    @property
    def _size(self):
        _size = _gl.GLint()
        with self:
            _gl.glGetBufferParameteriv(self._target, _gl.GL_BUFFER_SIZE, _gl.pointer(_size))
        return _size.value

    @property
    def usage(self):
        _usage = _gl.GLint()
        with self:
            _gl.glGetBufferParameteriv(self._target, _gl.GL_BUFFER_USAGE, _gl.pointer(_usage))
        return BaseBuffer.usages[_usage.value]

__all__ = ["BaseBuffer"]




import numpy

from glitter import Texture3D, Datatype, EnumConstant, Texture, uint8, int8, uint16, int16, uint32, int32, float32

def check_texture(shape, dtype, vrange):
    minval, maxval = vrange
    data = ((maxval - minval) * numpy.random.random(shape) + minval).astype(dtype.as_numpy())
    texture = Texture3D(data)
    assert (texture.data == data).all(), "data is broken"
    assert texture.shape == data.shape, "shape is broken (was %s, is %s)" % (data.shape, texture.shape)
    assert texture.dtype == Datatype.from_numpy(data.dtype), "dtype is broken (was %s, is %s)" % (Datatype.from_numpy(data.dtype), texture.dtype)

def test_texture_generator():
    shapes = ((4, 4, 4, 4), (4, 4, 4, 3), (4, 16, 8, 3), (5, 4, 4, 3), (5, 5, 5, 3), (6, 6, 6, 3), (7, 13, 5, 3), (1, 1, 3, 3))
    dtypes = (uint8, int8, uint16, int16, uint32, int32, float32)
    vranges = ((0, (1<<8)-1), (-1<<7, (1<<7)-1), (0, (1<<16)-1), (-1<<15, (1<<15)-1), (0, (1<<32)-1), (-1<<31, (1<<31)-1), (-10.0, 10.0))

    for shape in shapes:
        for dtype, vrange in zip(dtypes, vranges):
            yield check_texture, shape, dtype, vrange

def check_property(texture, name):
    value = getattr(texture, name)
    if isinstance(value, EnumConstant):
        valid_values = list(value._enum._reverse_dict.values())
        for value in valid_values:
            setattr(texture, name, value)
            assert getattr(texture, name) == value, "property %s is broken" % name
    else:
        setattr(texture, name, value)
        assert getattr(texture, name) == value, "property %s is broken" % name

def test_property_generator():
    texture = Texture3D(numpy.random.random((1, 1, 1, 4)).astype(numpy.float32))
    properties = [x for x in dir(texture) if x != "sampler" and not x.startswith("_") and type(getattr(Texture, x)) == property]

    for p in properties:
        if p in ("data", "shape", "dtype"):
            continue
        if getattr(Texture, p).fset is None:
            continue
        yield check_property, texture, p




"""Sync classes.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, client_wait_sync_returns

class Sync(ManagedObject):
    """@todo: Implement properties using C{glGetSynciv}."""
    _delete_id = _gl.glDeleteSync

    def __init__(self, context=None):
        super(Sync, self).__init__(context=context)

class FenceSync(Sync):
    _generate_id = lambda: _gl.glFenceSync(_gl.GL_SYNC_GPU_COMMANDS_COMPLETE, 0)

    client_wait_sync_returns = client_wait_sync_returns

    def wait(self):
        _gl.glWaitSync(self._id, 0, _gl.GL_TIMEOUT_IGNORED)

    def client_wait(self):
        return self.client_wait_sync_returns[_gl.glClientWaitSync(self._id, _gl.GL_SYNC_FLUSH_COMMANDS_BIT, _gl.GL_TIMEOUT_IGNORED)]

__all__ = ["Sync", "FenceSync"]




#!/usr/bin/env python

#! This file is a literate Python program. You can compile the documentation
#! using mylit (http://pypi.python.org/pypi/mylit/).
## title = "glitter Example: Introduction"
## stylesheet = "pygments_style.css"

# <h1><i>glitter</i> Example: Introduction</h1>

# <h2>Summary</h2>

# This program will open a GLUT window and render a textured, rotating fan,
# overlaid with a textured background. To illustrate the use of framebuffer
# objects, the scene will first be rendered to a texture, then displayed on
# screen by a different shader. This example has been adapted from <a
# href="http://openglbook.com/the-book/">OpenGLBook.com</a>.

# This is a fairly advanced "introductory" example using vertex arrays, shader
# programs, pipelines, textures, and logging. If you are looking for something
# simpler (but less complete) to start with, look at the <a
# href="simple.html">simple example</a>.

# <img src="introduction.png">

# <h2>Front matter</h2>

# <h3>Module docstring</h3>

# The module docstring is used as a description of this example in the
# generated documentation:
"""Basic example using L{VertexArray}s, L{ShaderProgram}s, L{Pipeline}s, L{Texture}s, and logging.

@author: Stephan Wenger
@date: 2012-02-29
"""

# <h3>Imports</h3>

# <i>glitter</i> can use the <a
# href="http://docs.python.org/library/logging.html">logging</a> module for
# logging all OpenGL commands issued by the library. The log entries are
# emitted at <i>DEBUG</i> level, so we enable printing of debug messages:
import logging
logging.basicConfig(level=logging.DEBUG)

# Our scene is going to rotate. The rotating modelview matrix is computed using
# the sine and cosine functions from the math module:
from math import sin, cos, pi

# <i>glitter</i> uses <a href="http://numpy.scipy.org/">numpy</a> for
# representation of array data. We will use numpy's <code>random()</code>
# function to generate random textures:
from numpy.random import random

# We can usually import classes and functions contained in <i>glitter</i>
# submodules directly from glitter:
from glitter import ShaderProgram, RectangleTexture, Texture2D, Pipeline, VertexArray, add_logger

# Modules with external dependencies other than numpy, such as platform
# dependent parts like methods for the generation of an OpenGL context,
# however, have to be imported from their respective submodules:
from glitter.contexts.glut import GlutWindow, main_loop, get_elapsed_time

# <h2>Shaders</h2>

# <h3>Primary shader program</h3>

# In the OpenGL core profile, there is no such thing as a "standard pipeline"
# any more. We can either use the minimalistic <code>defaultpipeline</code>
# from the <code>glitter.convenience</code> module, or define our own shaders.
# We'll do the latter since we want some additional functionality. Instead of
# loading the shader from a file at runtime, we can define it inline as a
# Python string:
vertex_shader = """
#version 400 core

layout(location=0) in vec4 in_position;
layout(location=1) in vec4 in_color;
uniform mat4 modelview_matrix;
out vec4 ex_color;
out vec2 ex_texcoord;

void main() {
    gl_Position = modelview_matrix * in_position;
    ex_color = in_color;
    ex_texcoord = in_position.xy * 0.5 + 0.5;
}
"""
# The vertex shader will receive its per-vertex inputs <code>in_position</code>
# and <code>in_color</code> from a vertex array. The uniform variable
# <code>modelview_matrix</code> will be set directly on the shader program
# object. The varying values <code>ex_color</code> and <code>ex_texcoord</code>
# are passed to the fragment shader.

# The fragment shader is executed once for every rasterized fragment. It
# receives interpolated values for <code>ex_color</code> and
# <code>ex_texcoord</code> from the vertex shader, reads from textures
# <code>texture_0</code> and <code>texture_1</code>, and writes the fragment
# color to <code>out_color</code>, which is written to a texture by the
# framebuffer object.
fragment_shader = """
#version 400 core
#extension GL_ARB_texture_rectangle : enable

in vec4 ex_color;
in vec2 ex_texcoord;
uniform sampler2D texture_0;
uniform sampler2DRect texture_1;
layout(location=0) out vec4 out_color;

void main() {
    out_color = 0.5 * ex_color
    + texture(texture_0, ex_texcoord)
    * texture(texture_1, gl_FragCoord.xy / 10.0)
    ;
}
"""
# The textures that are used by the fragment shader are automatically bound to
# free texture units by the shader object.

# <h3>Copying shader program</h3>

# To copy the image from a texture to the screen, we will use the following simple shaders:
copy_vertex_shader = """
#version 400 core

layout(location=0) in vec4 in_position;

void main() {
    gl_Position = in_position;
}
"""

copy_fragment_shader = """
#version 400 core
#extension GL_ARB_texture_rectangle : enable

uniform sampler2DRect image;
layout(location=0) out vec4 out_color;

void main() {
    out_color = texture(image, gl_FragCoord.xy);
}
"""

# <h2>Vertex arrays</h2>

# The geometry is specified as arrays (or nested lists) of vertices, colors,
# and indices into the vertex and color arrays. Lists are automatically
# converted to appropriate numpy arrays. The shape of these arrays describes
# how the arrays are to be drawn. When an index array is used, as in this
# example, the vertex and color arrays are two-dimensional: the first dimension
# is the number of vertices, the second is the number of values for each vertex
# (e.g. the red, green, blue, and alpha channels for a color array). The shape
# of the index array is also two-dimensional: the first dimension is the number
# of primitives to be drawn, the second is the number of vertices in each
# primitive. Here, we draw sixteen triangles.
vertices = (
    ( 0.0,  0.0, 0.0, 1.0), # center
    (-0.2,  0.8, 0.0, 1.0), ( 0.2,  0.8, 0.0, 1.0), ( 0.0,  0.8, 0.0, 1.0), ( 0.0,  1.0, 0.0, 1.0), # top
    (-0.2, -0.8, 0.0, 1.0), ( 0.2, -0.8, 0.0, 1.0), ( 0.0, -0.8, 0.0, 1.0), ( 0.0, -1.0, 0.0, 1.0), # bottom
    (-0.8, -0.2, 0.0, 1.0), (-0.8,  0.2, 0.0, 1.0), (-0.8,  0.0, 0.0, 1.0), (-1.0,  0.0, 0.0, 1.0), # left
    ( 0.8, -0.2, 0.0, 1.0), ( 0.8,  0.2, 0.0, 1.0), ( 0.8,  0.0, 0.0, 1.0), ( 1.0,  0.0, 0.0, 1.0), # right
    )

colors = (
    (1.0, 1.0, 1.0, 1.0), # center
    (0.0, 1.0, 0.0, 1.0), (0.0, 0.0, 1.0, 1.0), (0.0, 1.0, 1.0, 1.0), (1.0, 0.0, 0.0, 1.0), # top
    (0.0, 0.0, 1.0, 1.0), (0.0, 1.0, 0.0, 1.0), (0.0, 1.0, 1.0, 1.0), (1.0, 0.0, 0.0, 1.0), # bottom
    (0.0, 1.0, 0.0, 1.0), (0.0, 0.0, 1.0, 1.0), (0.0, 1.0, 1.0, 1.0), (1.0, 0.0, 0.0, 1.0), # left
    (0.0, 0.0, 1.0, 1.0), (0.0, 1.0, 0.0, 1.0), (0.0, 1.0, 1.0, 1.0), (1.0, 0.0, 0.0, 1.0), # right
    )

indices = (
    (0,  1,  3), (0,  3,  2), ( 3,  1,  4), ( 3,  4,  2), # top
    (0,  5,  7), (0,  7,  6), ( 7,  5,  8), ( 7,  8,  6), # bottom
    (0,  9, 11), (0, 11, 10), (11,  9, 12), (11, 12, 10), # left
    (0, 13, 15), (0, 15, 14), (15, 13, 16), (15, 16, 14), # right
    )
# If no index array is given, the vertex and color arrays are
# three-dimensional: the first dimension is then the number of primitives, the
# second is the number of vertices per primitive, and the third is the number
# of values per vertex. Since in a typical geometry several primitives share a
# common vertex, it is usually more memory efficient to use an index array.

# <h2>Main class</h2>

# We wrap all the OpenGL interaction in a class. The class will contain an
# <code>__init__()</code> method to set up all OpenGL objects, any required
# callback methods, as well as a <code>run()</code> method to trigger execution
# of the GLUT main loop.
class IntroductionExample(object):
    # <h3>Initialization</h3>

    # When an <code>IntroductionExample</code> instance is created, we need to
    # initialize a few OpenGL objects.
    def __init__(self):
        # First, we create a window; this also creates an OpenGL context.
        self.window = GlutWindow(double=True, multisample=True)

        # Then, we set the GLUT display and keyboard callback functions which
        # will be defined later.
        self.window.display_callback = self.display
        self.window.keyboard_callback = self.keyboard

        # A shader program is built from the previously defined vertex and
        # fragment codes:
        self.shader = ShaderProgram(vertex=vertex_shader, fragment=fragment_shader)

        # This shader program is then used to build a pipeline. A pipeline is a
        # convenience object that encapsulates a vertex array for input, a
        # shader program for processing, and a framebuffer for output. The
        # framebuffer is optional (we could render directly to the screen if we
        # wanted), but we will render into a texture and copy the texture to
        # the screen for instructional purposes. The <code>Pipeline</code>
        # constructor automatically creates an empty vertex array and a
        # framebuffer with no attachments. Named constructor arguments are
        # interpreted as attributes of the vertex array and the framebuffer or
        # as named inputs and outputs of the shader. This means we can directly
        # pass in arrays of vertices and colors that will be bound to
        # <code>in_position</code> and <code>in_color</code>, respectivey, as
        # well as the array of element indices to draw and an empty texture to
        # bind to the framebuffer:
        self.render_pipeline = Pipeline(self.shader, in_position=vertices, in_color=colors,
                elements=indices, out_color=RectangleTexture(shape=(300, 300, 3)))

        # Shader uniform variables like textures can also be set directly on
        # the pipeline. Here we initialize two textures with random data:
        self.render_pipeline.texture_0 = Texture2D(random((30, 30, 4)))
        self.render_pipeline.texture_1 = RectangleTexture(random((30, 30, 4)))

        # Many properties, such as the filtering mode for textures, can
        # directly be set as attributes on the corresponding objects:
        self.render_pipeline.texture_0.min_filter = Texture2D.min_filters.NEAREST
        self.render_pipeline.texture_0.mag_filter = Texture2D.mag_filters.NEAREST

        # For copying the texture to the screen, we create another shader
        # program.
        self.copy_shader = ShaderProgram(vertex=copy_vertex_shader, fragment=copy_fragment_shader)

        # The input texture of this shader program is the output texture of the
        # previous pipeline. Since all textures and framebuffers are
        # automatically bound and unbound, we do not need to worry about
        # whether the framebuffer is still writing to the texture.
        self.copy_shader.image = self.render_pipeline.out_color

        # Instead of using a pipeline with named vertex shader inputs, we can
        # also create a vertex array object directly with a list of vertex
        # shader inputs to use. Here we use only a single vertex shader input:
        # the coordinates of a fullscreen quad. The <code>elements</code>
        # parameter defines two triangles that make up the quad.
        self.vao = VertexArray(((-1.0, -1.0), (-1.0, 1.0), (1.0, 1.0), (1.0, -1.0)), elements=((0, 1, 2), (0, 2, 3)))

    # <h3>Callback functions</h3>

    # <h4>Display function</h4>

    # Here we define the display function. It will be called by GLUT whenever
    # the screen has to be redrawn.
    def display(self):
        # We can simply clear the pipeline and render the vertices into the
        # framebuffer using the shader with the following two lines:
        self.render_pipeline.clear()
        self.render_pipeline.draw()

        # For output on the screen, we have created a GLUT window. It can be
        # cleared in very much the same way:
        self.window.clear()

        # To copy the results of the pipeline to the screen, we use the shader
        # that simply displays a texture. The shader can be bound by using a
        # <code>with</code> statement:
        with self.copy_shader:
            # All textures used by the shader are then bound automatically, and
            # everything is reset to its previous state when we leave the
            # <code>with</code> block.

            # With the shader bound, we simply draw a fullscreen quad that is
            # stored in a vertex array we will create in the initialization
            # section:
            self.vao.draw()

        # After all rendering commands have been issued, we swap the back
        # buffer to the front, making the rendered image visible all at once:
        self.window.swap_buffers()

        # Finally, we disable logging so that we only see the OpenGL calls of
        # the first run of the display function:
        add_logger(None)

    # <h4>Keyboard function</h4>

    # To further illustrate the concept of GLUT callbacks, here's a keyboard
    # handler that will simply make the program exit when any key is pressed:
    def keyboard(self, key, x, y):
        raise SystemExit

    # <h4>Timer function</h4>

    # The animation is controlled by a GLUT timer. The timer callback changes
    # the modelview matrix, schedules the next timer event, and causes a screen
    # redraw:
    def timer(self):
        # We first get the elapsed time from GLUT using
        # <code>get_elapsed_time()</code>:
        t = get_elapsed_time()
        phi = 2 * pi * t / 4.0

        # We then set the <code>modelview_matrix</code> uniform variable of the
        # shader created in the initialization section simply by setting an
        # attribute:
        self.shader.modelview_matrix = ((cos(phi), sin(phi), 0, 0), (-sin(phi), cos(phi), 0, 0), (0, 0, 1, 0), (0, 0, 0, 1))

        # The following line schedules the next timer event to execute after
        # ten milliseconds.
        self.window.add_timer(10, self.timer)

        # Finally, we tell GLUT to redraw the screen.
        self.window.post_redisplay()

    # <h3>Running</h3>
    
    # We will call the <code>run()</code> method later to run the OpenGL code.
    def run(self):
        # To start the animation, we call the timer once; all subsequent timer
        # calls will be scheduled by the timer function itself.
        self.timer()

        # Now that all the initialization is done, we add the default logger to
        # all OpenGL commands so that we can see what OpenGL the display
        # function issues, and in which order.
        add_logger()

        # Finally, to start rendering, we enter the GLUT main loop.
        main_loop()

    # When the main loop exits, control is handed back to the script, unless
    # <code>SystemExit</code> has been raised by the keyboard handler.

# <h2>Main section</h2>

# Finally, if this program is being run from the command line, we instanciate
# the main class and run it.
if __name__ == "__main__":
    IntroductionExample().run()




#!/usr/bin/env python

#! This file is a literate Python program. You can compile the documentation
#! using mylit (http://pypi.python.org/pypi/mylit/).
## title = "glitter Example: Mesh Viewer"
## stylesheet = "pygments_style.css"

# <h1><i>glitter</i> Example: Mesh Viewer</h1>

# <h2>Summary</h2>

# This program will open a GLUT window and render a mesh from an <a
# href="www.hdfgroup.org/HDF5/">HDF5</a> data file.

# <img src="meshview.png">

# <h2>Front matter</h2>

# <h3>Module docstring</h3>

# The module docstring is used as a description of this example in the
# generated documentation:
"""Simple mesh viewer.

@author: Stephan Wenger
@date: 2012-02-29
"""

# <h3>Imports</h3>

# Our scene is going to rotate. The rotating modelview matrix is computed using
# the sine and cosine functions from the math module:
from math import sin, cos, pi

# <i>glitter</i> uses <a href="http://numpy.scipy.org/">numpy</a> for
# representation of array data. We will use numpy's <code>random()</code>
# function to generate random textures:
from numpy.random import random

# We assume the mesh is stored in a <a
# href="http://www.hdfgroup.org/HDF5/">HDF5</a> file, so import <a
# href="h5py.alfven.org"><code>h5py</code></a>.
import h5py

# We can usually import classes and functions contained in <i>glitter</i>
# submodules directly from glitter:
from glitter import VertexArray, State, get_default_program

# Modules with external dependencies other than numpy, such as platform
# dependent parts like methods for the generation of an OpenGL context,
# however, have to be imported from their respective submodules:
from glitter.contexts.glut import GlutWindow, main_loop, get_elapsed_time

# <h2>Main class</h2>

# We wrap all the OpenGL interaction in a class. The class will contain an
# <code>__init__()</code> method to set up all OpenGL objects, any required
# callback methods, as well as a <code>run()</code> method to trigger execution
# of the GLUT main loop.
class MeshViewer(object):
    # <h3>Initialization</h3>

    # When a <code>MeshViewer</code> instance is created, we need to
    # initialize a few OpenGL objects.
    def __init__(self, filename):
        # First, we create a window; this also creates an OpenGL context.
        self.window = GlutWindow(double=True, multisample=True)

        # Then, we set the GLUT display callback function which will be defined later.
        self.window.display_callback = self.display

        # In the OpenGL core profile, there is no such thing as a "standard pipeline"
        # any more. We use the minimalistic <code>defaultpipeline</code> from the
        # <code>glitter.convenience</code> module to create a shader program instead:
        self.shader = get_default_program()

        # We open the HDF5 file specified on the command line for reading:
        with h5py.File(filename, "r") as f:
            # The vertices, colors and indices of the mesh are read from the
            # corresponding datasets in the HDF5 file. Note that the names of the
            # datasets are mere convention. Colors and indices are allowed to be
            # undefined.
            vertices = f["vertices"]
            colors = f.get("colors", None)
            elements = f.get("indices", None)

            # If no colors were specified, we generate random ones so we can
            # distinguish the triangles without fancy shading.
            if colors is None:
                colors = random((len(vertices), 3))[:, None, :][:, [0] * vertices.shape[1], :]

            # Here, we create a vertex array that contains buffers for two vertex array
            # input variables as well as an index array. If <code>elements</code>
            # is <code>None</code>, the vertex array class will draw all vertices
            # in order.
            self.vao = VertexArray(vertices, colors, elements=elements)

    # <h3>Callback functions</h3>

    # <h4>Display function</h4>

    # Here we define the display function. It will be called by GLUT whenever the
    # screen has to be redrawn.
    def display(self):
        # First we clear the default framebuffer:
        self.window.clear()

        # To draw the vertex array, we use:
        self.vao.draw()

        # After all rendering commands have been issued, we swap the back buffer to
        # the front, making the rendered image visible all at once:
        self.window.swap_buffers()

    # <h4>Timer function</h4>

    # The animation is controlled by a GLUT timer. The timer callback changes the
    # modelview matrix, schedules the next timer event, and causes a screen redraw:
    def timer(self):
        # We first get the elapsed time from GLUT using <code>get_elapsed_time()</code>:
        phi = 2 * pi * get_elapsed_time() / 20.0

        # We then set the <code>modelview_matrix</code> uniform variable of the
        # shader simply by setting an attribute:
        self.shader.modelview_matrix = ((cos(phi), 0, sin(phi), 0), (0, 1, 0, 0), (-sin(phi), 0, cos(phi), 0), (0, 0, 0, 1))

        # The following line schedules the next timer event to execute after ten milliseconds.
        self.window.add_timer(10, self.timer)

        # Finally, we tell GLUT to redraw the screen.
        self.window.post_redisplay()

    # <h3>Running</h3>
    
    # We will call the <code>run()</code> method later to run the OpenGL code.
    def run(self):
        # To start the animation, we call the timer once; all subsequent timer
        # calls will be scheduled by the timer function itself.
        self.timer()

        # The shader program is bound by using a <code>with</code> statement:
        with self.shader:
            # The <code>State</code> class encapsulates state changes in the
            # context. For example, to enable depth testing for the duration of the
            # following function call, we would write:
            with State(depth_test=True):
                # With the shader bound and depth testing enabled, we enter the
                # GLUT main loop.
                main_loop()

        # When the main loop exits, control is handed back to the script.

# <h2>Main section</h2>

# Finally, if this program is being run from the command line, we instanciate
# the main class and run it.
if __name__ == "__main__":
    import sys
    MeshViewer(sys.argv[1]).run()




"""Texture classes.

@todo: Check OpenGL memory layout; do shaders use the same coordinates as C{numpy} does?
@todo: Implement C{__getitem__} and C{__setitem__} for subimages (C{glTexSubImage3D}, C{glGetTexImage} with C{format=GL_RED} etc.).
@todo: Implement stencil textures.
@todo: Implement image textures.
@todo: Implement normalized integer types (GL_RGBA8_SNORM etc.)

@author: Stephan Wenger
@date: 2012-02-29
"""

import numpy as _np

import glitter.raw as _gl
from glitter.utils import texture_compare_funcs, texture_compare_modes, texture_min_filters, texture_mag_filters, texture_swizzles, texture_wrapmodes, dtype_to_gl_iformat, dtype_to_gl_format, gl_iformat_to_dtype, gl_iformat_to_gl_type, Datatype, coerce_array, ManagedObject, BindReleaseObject, float32, texture_formats

class Texture(ManagedObject, BindReleaseObject):
    _generate_id = _gl.glGenTextures
    _delete_id = _gl.glDeleteTextures
    _db = "textures"

    _ndim = NotImplemented
    _set = NotImplemented

    compare_funcs = texture_compare_funcs
    compare_modes = texture_compare_modes
    min_filters = texture_min_filters
    mag_filters = texture_mag_filters
    swizzles = texture_swizzles
    wrapmodes = texture_wrapmodes

    def bind(self):
        return self._context.texture_units.bind(self)

    def release(self):
        self._context.texture_units.release(self)

    def __init__(self, data=None, shape=None, dtype=None, depth=False, stencil=False, mipmap=False, context=None, **kwargs):
        if any(x is NotImplemented for x in (self._ndim, self._set)):
            raise TypeError("%s is abstract" % self.__class__.__name__)
        super(Texture, self).__init__(context=context)
        self.set_data(data, shape, dtype, depth=depth, stencil=stencil, mipmap=mipmap)
        for key, value in list(kwargs.items()):
            setattr(self, key, value)

    def __getitem__(self, key):
        """Return a tuple describing a texture layer.

        This is mainly provided as a convenience notation for binding texture
        layers to L{Framebuffer}s.
        """

        return (self, key)

    def __len__(self):
        return self.shape[0]

    def set_data(self, data=None, shape=None, dtype=None, level=0, depth=False, stencil=False, mipmap=False):
        if data is None:
            if shape is None:
                raise ValueError("must specify either data or shape")
            if dtype is None:
                dtype = float32
        else:
            data = coerce_array(data, dtype)
            if shape is not None:
                data = data.reshape(shape)
            shape = data.shape
            dtype = Datatype.from_numpy(data.dtype)

        if len(shape) != self._ndim:
            raise TypeError("shape must be %d-dimensional" % self._ndim)

        _iformat = dtype_to_gl_iformat[dtype, shape[-1]]
        _format = dtype_to_gl_format[dtype, shape[-1]]
        _type = dtype._as_gl()

        if depth and stencil:
            raise NotImplementedError("depth-and-stencil textures are currently not supported")
        elif depth:
            _iformat = _format = None
            for (nptype, colors), ifmt, (gltype, fmt) in texture_formats:
                if nptype == dtype and colors == shape[-1] and fmt == _gl.GL_DEPTH_COMPONENT:
                    _iformat, _format = ifmt, fmt
            if _iformat is None or _format is None:
                raise TypeError("no matching depth texture format for %d %s colors" % (shape[-1], dtype))
        elif stencil:
            raise NotImplementedError("stencil textures are currently not supported")
        
        _data = data.ctypes if data is not None else _gl.POINTER(_gl.GLvoid)()
        _gl.glPixelStorei(_gl.GL_UNPACK_ALIGNMENT, 1)
        with self:
            args = [self._target, level, _iformat] + list(reversed(shape[:-1])) + [0, _format, _type, _data]
            self._set(*args)
        if dtype.is_float():
            self.min_filter = self.min_filters.LINEAR_MIPMAP_LINEAR if mipmap else self.min_filters.LINEAR
            self.mag_filter = self.mag_filters.LINEAR
        else:
            self.min_filter = self.min_filters.NEAREST
            self.mag_filter = self.mag_filters.NEAREST

        if mipmap:
            self.generate_mipmap()

    def get_data(self, level=0):
        _data = _np.empty(self.shape, dtype=self.dtype.as_numpy())
        _gl.glPixelStorei(_gl.GL_PACK_ALIGNMENT, 1)
        with self:
            _gl.glGetTexImage(self._target, level, self._format, self._type, _data.ctypes)
        return _data

    def generate_mipmap(self):
        with self:
            _gl.glGenerateMipmap(self._target)

    @property
    def data(self):
        return self.get_data()

    @data.setter
    def data(self, data):
        self.set_data(data)

    @property
    def shape(self):
        with self:
            colors = gl_iformat_to_dtype[self._iformat][1]        
            _width = _gl.GLint()
            _gl.glGetTexLevelParameteriv(self._target, 0, _gl.GL_TEXTURE_WIDTH, _gl.pointer(_width))
            if self._ndim == 2:
                return (_width.value, colors)
            _height = _gl.GLint()
            _gl.glGetTexLevelParameteriv(self._target, 0, _gl.GL_TEXTURE_HEIGHT, _gl.pointer(_height))
            if self._ndim == 3:
                return (_height.value, _width.value, colors)
            _depth = _gl.GLint()
            _gl.glGetTexLevelParameteriv(self._target, 0, _gl.GL_TEXTURE_DEPTH, _gl.pointer(_depth))
            if self._ndim == 4:
                return (_depth.value, _height.value, _width.value, colors)

    @property
    def dtype(self):
        return gl_iformat_to_dtype[self._iformat][0]

    @property
    def _iformat(self):
        _iformat = _gl.GLint()
        with self:
            _gl.glGetTexLevelParameteriv(self._target, 0, _gl.GL_TEXTURE_INTERNAL_FORMAT, _gl.pointer(_iformat))
        return _iformat.value

    @property
    def _format(self):
        return dtype_to_gl_format[self.dtype, self.shape[-1]]

    @property
    def _type(self):
        return gl_iformat_to_gl_type[self._iformat]

    @property
    def base_level(self):
        _base_level = _gl.GLint()
        with self:
            _gl.glGetTexParameterIiv(self._target, _gl.GL_TEXTURE_BASE_LEVEL, _gl.pointer(_base_level))
        return _base_level.value

    @base_level.setter
    def base_level(self, base_level):
        with self:
            _gl.glTexParameterIiv(self._target, _gl.GL_TEXTURE_BASE_LEVEL, _gl.pointer(_gl.GLint(base_level)))

    @property
    def border_color(self):
        if self.dtype.is_float():
            _border_color = (_gl.GLfloat * 4)()
            with self:
                _gl.glGetTexParameterfv(self._target, _gl.GL_TEXTURE_BORDER_COLOR, _border_color)
        elif self.dtype.is_signed():
            _border_color = (_gl.GLint * 4)()
            with self:
                _gl.glGetTexParameterIiv(self._target, _gl.GL_TEXTURE_BORDER_COLOR, _border_color)
        else:
            _border_color = (_gl.GLuint * 4)()
            with self:
                _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_BORDER_COLOR, _border_color)
        return [_border_color[i] for i in range(4)]

    @border_color.setter
    def border_color(self, border_color):
        if self.dtype.is_float():
            _border_color = (_gl.GLfloat * 4)()
            for i, v in zip(list(range(4)), border_color):
                _border_color[i] = v
            with self:
                _gl.glTexParameterfv(self._target, _gl.GL_TEXTURE_BORDER_COLOR, _border_color)
        elif self.dtype.is_signed():
            _border_color = (_gl.GLint * 4)()
            for i, v in zip(list(range(4)), border_color):
                _border_color[i] = v
            with self:
                _gl.glTexParameterIiv(self._target, _gl.GL_TEXTURE_BORDER_COLOR, _border_color)
        else:
            _border_color = (_gl.GLuint * 4)()
            for i, v in zip(list(range(4)), border_color):
                _border_color[i] = v
            with self:
                _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_BORDER_COLOR, _border_color)

    @property
    def compare_func(self):
        _compare_func = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_COMPARE_FUNC, _gl.pointer(_compare_func))
        return self.compare_funcs[_compare_func.value]

    @compare_func.setter
    def compare_func(self, compare_func):
        _compare_func = _gl.GLenum(self.compare_funcs(compare_func)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_COMPARE_FUNC, _gl.pointer(_compare_func))

    @property
    def compare_mode(self):
        _compare_mode = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_COMPARE_MODE, _gl.pointer(_compare_mode))
        return self.compare_modes[_compare_mode.value]

    @compare_mode.setter
    def compare_mode(self, compare_mode):
        _compare_mode = _gl.GLenum(self.compare_modes(compare_mode)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_COMPARE_MODE, _gl.pointer(_compare_mode))
    
    @property
    def immutable_format(self): # Textures become immutable if their storage is specified with glTexStorage1D, glTexStorage2D or glTexStorage3D
        _immutable_format = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_COMPARE_MODE, _gl.pointer(_immutable_format))
        return bool(_immutable_format.value)

    @property
    def lod_bias(self):
        _lod_bias = _gl.GLfloat()
        with self:
            _gl.glGetTexParameterfv(self._target, _gl.GL_TEXTURE_LOD_BIAS, _gl.pointer(_lod_bias))
        return _lod_bias.value

    @lod_bias.setter
    def lod_bias(self, lod_bias):
        with self:
            _gl.glTexParameterfv(self._target, _gl.GL_TEXTURE_LOD_BIAS, _gl.pointer(_gl.GLfloat(lod_bias)))

    @property
    def min_filter(self):
        _min_filter = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_MIN_FILTER, _gl.pointer(_min_filter))
        return self.min_filters[_min_filter.value]

    @min_filter.setter
    def min_filter(self, min_filter):
        _min_filter = _gl.GLenum(self.min_filters(min_filter)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_MIN_FILTER, _gl.pointer(_min_filter))

    @property
    def mag_filter(self):
        _mag_filter = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_MAG_FILTER, _gl.pointer(_mag_filter))
        return self.mag_filters[_mag_filter.value]

    @mag_filter.setter
    def mag_filter(self, mag_filter):
        _mag_filter = _gl.GLenum(self.mag_filters(mag_filter)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_MAG_FILTER, _gl.pointer(_mag_filter))

    @property
    def min_lod(self):
        _min_lod = _gl.GLint()
        with self:
            _gl.glGetTexParameterIiv(self._target, _gl.GL_TEXTURE_MIN_LOD, _gl.pointer(_min_lod))
        return _min_lod.value

    @min_lod.setter
    def min_lod(self, min_lod):
        with self:
            _gl.glTexParameterIiv(self._target, _gl.GL_TEXTURE_MIN_LOD, _gl.pointer(_gl.GLint(min_lod)))

    @property
    def max_lod(self):
        _max_lod = _gl.GLint()
        with self:
            _gl.glGetTexParameterIiv(self._target, _gl.GL_TEXTURE_MAX_LOD, _gl.pointer(_max_lod))
        return _max_lod.value

    @max_lod.setter
    def max_lod(self, max_lod):
        with self:
            _gl.glTexParameterIiv(self._target, _gl.GL_TEXTURE_MAX_LOD, _gl.pointer(_gl.GLint(max_lod)))

    @property
    def max_level(self):
        _max_level = _gl.GLint()
        with self:
            _gl.glGetTexParameterIiv(self._target, _gl.GL_TEXTURE_MAX_LEVEL, _gl.pointer(_max_level))
        return _max_level.value

    @max_level.setter
    def max_level(self, max_level):
        with self:
            _gl.glTexParameterIiv(self._target, _gl.GL_TEXTURE_MAX_LEVEL, _gl.pointer(_gl.GLint(max_level)))

    @property
    def swizzle_r(self):
        _swizzle_r = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_R, _gl.pointer(_swizzle_r))
        return self.swizzles[_swizzle_r.value]

    @swizzle_r.setter
    def swizzle_r(self, swizzle_r):
        _swizzle_r = _gl.GLenum(self.swizzles(swizzle_r)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_R, _gl.pointer(_swizzle_r))

    @property
    def swizzle_g(self):
        _swizzle_g = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_G, _gl.pointer(_swizzle_g))
        return self.swizzles[_swizzle_g.value]

    @swizzle_g.setter
    def swizzle_g(self, swizzle_g):
        _swizzle_g = _gl.GLenum(self.swizzles(swizzle_g)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_G, _gl.pointer(_swizzle_g))

    @property
    def swizzle_b(self):
        _swizzle_b = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_B, _gl.pointer(_swizzle_b))
        return self.swizzles[_swizzle_b.value]

    @swizzle_b.setter
    def swizzle_b(self, swizzle_b):
        _swizzle_b = _gl.GLenum(self.swizzles(swizzle_b)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_B, _gl.pointer(_swizzle_b))

    @property
    def swizzle_a(self):
        _swizzle_a = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_A, _gl.pointer(_swizzle_a))
        return self.swizzles[_swizzle_a.value]

    @swizzle_a.setter
    def swizzle_a(self, swizzle_a):
        _swizzle_a = _gl.GLenum(self.swizzles(swizzle_a)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_A, _gl.pointer(_swizzle_a))

    @property
    def swizzle_rgba(self):
        _swizzle_rgba = (_gl.GLenum * 4)()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_RGBA, _swizzle_rgba)
        return [self.swizzles[_swizzle_rgba[i]] for i in range(4)]

    @swizzle_rgba.setter
    def swizzle_rgba(self, swizzle_rgba):
        _swizzle_rgba = (_gl.GLenum * 4)()
        for i, v in zip(list(range(4)), swizzle_rgba):
            _swizzle_rgba[i] = self.swizzles(v)._value
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_SWIZZLE_RGBA, _swizzle_rgba)

    @property
    def wrap_s(self):
        _wrap_s = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_WRAP_S, _gl.pointer(_wrap_s))
        return self.wrapmodes[_wrap_s.value]

    @wrap_s.setter
    def wrap_s(self, wrap_s):
        _wrap_s = _gl.GLenum(self.wrapmodes(wrap_s)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_WRAP_S, _gl.pointer(_wrap_s))

    @property
    def wrap_t(self):
        _wrap_t = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_WRAP_T, _gl.pointer(_wrap_t))
        return self.wrapmodes[_wrap_t.value]

    @wrap_t.setter
    def wrap_t(self, wrap_t):
        _wrap_t = _gl.GLenum(self.wrapmodes(wrap_t)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_WRAP_T, _gl.pointer(_wrap_t))

    @property
    def wrap_r(self):
        _wrap_r = _gl.GLenum()
        with self:
            _gl.glGetTexParameterIuiv(self._target, _gl.GL_TEXTURE_WRAP_R, _gl.pointer(_wrap_r))
        return self.wrapmodes[_wrap_r.value]

    @wrap_r.setter
    def wrap_r(self, wrap_r):
        _wrap_r = _gl.GLenum(self.wrapmodes(wrap_r)._value)
        with self:
            _gl.glTexParameterIuiv(self._target, _gl.GL_TEXTURE_WRAP_R, _gl.pointer(_wrap_r))

class Texture1D(Texture):
    _target = _gl.GL_TEXTURE_1D
    _binding = "texture_binding_1d"
    _ndim = 2
    _set = _gl.glTexImage1D

class Texture2D(Texture):
    _target = _gl.GL_TEXTURE_2D
    _binding = "texture_binding_2d"
    _ndim = 3
    _set = _gl.glTexImage2D

class TextureArray1D(Texture):
    _target = _gl.GL_TEXTURE_1D_ARRAY
    _binding = "texture_binding_1d_array"
    _ndim = 3
    _set = _gl.glTexImage2D

class RectangleTexture(Texture):
    _target = _gl.GL_TEXTURE_RECTANGLE
    _binding = "texture_binding_rectangle"
    _ndim = 3
    _set = _gl.glTexImage2D

class BufferTexture(Texture):
    """Texture that stores its data in a buffer.
    
    @todo: Override constructor, C{set_data} and C{get_data}; use C{glTexBuffer}
    """

    _target = _gl.GL_TEXTURE_BUFFER
    _binding = "texture_binding_buffer"
    _ndim = 3
    _set = _gl.glTexImage2D

class CubeMapTexture(Texture):
    _target = _gl.GL_TEXTURE_CUBE_MAP
    _binding = "texture_binding_cube_map"
    _ndim = 3
    _set = _gl.glTexImage2D

class MultisampleTexture2D(Texture):
    _target = _gl.GL_TEXTURE_2D_MULTISAMPLE
    _binding = "texture_binding_2d_multisample"
    _ndim = 3
    _set = _gl.glTexImage2D

class Texture3D(Texture):
    _target = _gl.GL_TEXTURE_3D
    _binding = "texture_binding_3d"
    _ndim = 4
    _set = _gl.glTexImage3D

class TextureArray2D(Texture):
    _target = _gl.GL_TEXTURE_2D_ARRAY
    _binding = "texture_binding_2d_array"
    _ndim = 4
    _set = _gl.glTexImage3D

class MultisampleTextureArray2D(Texture):
    _target = _gl.GL_TEXTURE_2D_MULTISAMPLE_ARRAY
    _binding = "texture_binding_2d_multisample_array"
    _ndim = 4
    _set = _gl.glTexImage3D

__all__ = [
    "Texture",
    "Texture1D",
    "Texture2D",
    "TextureArray1D",
    "RectangleTexture",
    "BufferTexture",
    "CubeMapTexture",
    "MultisampleTexture2D",
    "Texture3D",
    "TextureArray2D",
    "MultisampleTextureArray2D",
]




"""Shaders, shader programs, and program pipelines.

@author: Stephan Wenger
@date: 2012-02-29
"""

from glitter.shaders.programpipeline import *
from glitter.shaders.shader import *
from glitter.shaders.shaderprogram import *




"""GLUT context creation and management.

@bug: Closing a GLUT window via the window system (not the API) causes bogus error messages and even the occasional segfault.
@todo: Implement subwindows using C{glutCreateSubWindow} and C{glutGet(GLUT_WINDOW_PARENT)}.
@todo: Implement menus, font rendering, and geometric object rendering,
@todo: Fail gracefully (better error message) when context cannot be created, e.g. when opengl 3.2 not available
@todo: Key constants like C{GLUT_KEY_F1} should be available as an enum.

@author: Stephan Wenger
@date: 2012-02-29
"""

import random as _random

import glitter.raw as _gl
from glitter.utils import Enum
from glitter.contexts.context import Context
from glitter.contexts.contextmanager import context_manager

_cursors = Enum(
    right_arrow=_gl.GLUT_CURSOR_RIGHT_ARROW,
    left_arrow=_gl.GLUT_CURSOR_LEFT_ARROW,
    info=_gl.GLUT_CURSOR_INFO,
    destroy=_gl.GLUT_CURSOR_DESTROY,
    help=_gl.GLUT_CURSOR_HELP,
    cycle=_gl.GLUT_CURSOR_CYCLE,
    spray=_gl.GLUT_CURSOR_SPRAY,
    wait=_gl.GLUT_CURSOR_WAIT,
    text=_gl.GLUT_CURSOR_TEXT,
    crosshair=_gl.GLUT_CURSOR_CROSSHAIR,
    up_down=_gl.GLUT_CURSOR_UP_DOWN,
    left_right=_gl.GLUT_CURSOR_LEFT_RIGHT,
    top_side=_gl.GLUT_CURSOR_TOP_SIDE,
    bottom_side=_gl.GLUT_CURSOR_BOTTOM_SIDE,
    left_side=_gl.GLUT_CURSOR_LEFT_SIDE,
    right_side=_gl.GLUT_CURSOR_RIGHT_SIDE,
    top_left_corner=_gl.GLUT_CURSOR_TOP_LEFT_CORNER,
    top_right_corner=_gl.GLUT_CURSOR_TOP_RIGHT_CORNER,
    bottom_right_corner=_gl.GLUT_CURSOR_BOTTOM_RIGHT_CORNER,
    bottom_left_corner=_gl.GLUT_CURSOR_BOTTOM_LEFT_CORNER,
    full_crosshair=_gl.GLUT_CURSOR_FULL_CROSSHAIR,
    none=_gl.GLUT_CURSOR_NONE,
    inherit=_gl.GLUT_CURSOR_INHERIT,
)

def _func_property(glut_func):
    import re
    name = re.match("^glut(.*)Func$", glut_func.__name__).groups()[0]
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)).lower()
    private_name = "_%s_func" % name
    private_c_name = "_%s_func_c" % name

    def getter(self):
        if not hasattr(self, private_name):
            setattr(self, private_name, None)
        return getattr(self, private_name)

    def setter(self, func):
        ftype = glut_func.argtypes[0]
        setattr(self, private_c_name, ftype(func) if func is not None else ftype())
        setattr(self, private_name, func)
        with self:
            glut_func(getattr(self, private_c_name))

    return property(getter, setter)

def initialize(argv=None):
    """Initialize GLUT. Called automatically on import."""
    if argv is None:
        import sys
        argv = sys.argv
    _argc = _gl.c_int(len(argv))
    _argv = (_gl.c_char_p * _argc.value)()
    if argv is not None:
        for i, a in enumerate(argv):
            _argv[i] = a.encode("ascii") if isinstance(a, str) else a
    _gl.glutInit(_gl.pointer(_argc), _argv)
    argv[:] = [_argv[i] for i in range(_argc.value)]

def main_loop():
    """Enter the GLUT main loop.
    
    @todo: Avoid destroying a window in C{__del__} that has already been destroyed by closing it; can C{glutCloseFunc} help?
    """
    _gl.glutMainLoop()
    initialize()

def main_loop_event():
    """Allow GLUT to process events."""
    _gl.glutMainLoopEvent()

def leave_main_loop():
    """Leave the GLUT main loop."""
    _gl.glutLeaveMainLoop()

def get_screen_shape():
    """Screen height and width in pixels."""
    return _gl.glutGet(_gl.GLUT_SCREEN_HEIGHT, _gl.GLUT_SCREEN_WIDTH)

def get_screen_shape_mm():
    """Screen height and width in millimeters."""
    return _gl.glutGet(_gl.GLUT_SCREEN_HEIGHT_MM, _gl.GLUT_SCREEN_WIDTH_MM)

def get_elapsed_time():
    """Time in seconds since GLUT was initialized."""
    return _gl.glutGet(_gl.GLUT_ELAPSED_TIME) / 1000.0

def get_shift_state():
    """Only available in keyboard, special, and mouse callbacks."""
    return _gl.glutGetModifiers(_gl.GLUT_ACTIVE_SHIFT)

def get_ctrl_state():
    """Only available in keyboard, special, and mouse callbacks."""
    return _gl.glutGetModifiers(_gl.GLUT_ACTIVE_CTRL)

def get_alt_state():
    """Only available in keyboard, special, and mouse callbacks."""
    return _gl.glutGetModifiers(_gl.GLUT_ACTIVE_ALT)

class GlutWindow(Context):
    cursors = _cursors

    def __init__(self, version=(3, 2), compatibility_profile=False,
            debug=False, forward_compatible=True, shape=(300, 300),
            position=(-1, -1), index=False, double=False, accum=False,
            alpha=False, depth=False, stencil=False, multisample=False,
            stereo=False, luminance=False, name="", hide=False):
        """Create a new GLUT window.

        `version` is the OpenGL version to use. It can be either an integer or a (major, minor) tuple.
        If `compatibility_profile` is `True`, the compatibility profile will be used instead of the core profile.
        `debug` enables OpenGL debug mode, `forward_compatible` the forward compatibility mode.
        `shape` and `position` control the initial (height, width) shape and (y, x) position of the window.
        `index` enables color index mode instead of RGBA.
        `double` enables double buffering.
        `accum`, `alpha`, `depth`, and `stencil` enable the corresponding buffers.
        `multisample` enables multisampling.
        `stereo` enables stereo buffers.
        `luminance` enables luminance rendering instead of true RGBA.
        `name` is the name of the window and its initial title.
        If `hide` is `True`, the window will initially be hidden.
        """

        self._stack = []

        self._idle_func = None
        self._display_func = None
        self._reshape_func = None
        self._mouse_func = None
        self._motion_func = None
        self._keyboard_func = None
        
        self._timer_funcs = {}
        def timer_func(value):
            self._timer_funcs.pop(value)()
        self._timer_func = _gl.glutTimerFunc.argtypes[1](timer_func)

        self._name = self._window_title = self._icon_title = name

        if hasattr(version, "__iter__") and not isinstance(version, str):
            _gl.glutInitContextVersion(*version)
        else:
            _gl.glutInitContextVersion(version, 0)

        _gl.glutInitContextProfile(_gl.GLUT_COMPATIBILITY_PROFILE if compatibility_profile else _gl.GLUT_CORE_PROFILE)

        _gl.glutInitContextFlags(
                (_gl.GLUT_DEBUG if debug else 0) |
                (_gl.GLUT_FORWARD_COMPATIBLE if forward_compatible else 0)
                )

        height, width = shape
        _gl.glutInitWindowSize(width, height)

        y, x = position
        _gl.glutInitWindowPosition(x, y)

        _gl.glutInitDisplayMode(
                (_gl.GLUT_INDEX if index else _gl.GLUT_RGBA) |
                (_gl.GLUT_DOUBLE if double else _gl.GLUT_SINGLE) |
                (_gl.GLUT_ACCUM if accum else 0) |
                (_gl.GLUT_ALPHA if alpha else 0) |
                (_gl.GLUT_DEPTH if depth else 0) |
                (_gl.GLUT_STENCIL if stencil else 0) |
                (_gl.GLUT_MULTISAMPLE if multisample else 0) |
                (_gl.GLUT_STEREO if stereo else 0) |
                (_gl.GLUT_LUMINANCE if luminance else 0)
                )

        _gl.glutSetOption(_gl.GLUT_ACTION_ON_WINDOW_CLOSE, _gl.GLUT_ACTION_CONTINUE_EXECUTION)

        if not _gl.glutGet(_gl.GLUT_DISPLAY_MODE_POSSIBLE):
            raise RuntimeError("display mode not possible")

        old_binding = context_manager.current_context
        self._id = _gl.glutCreateWindow(self._name.encode("utf8")) # creating a window changes the current context without the context manager's knowledge
        if old_binding:
            old_binding._bind() # rebind the previous context circumventing any caching performed by the context
        else:
            context_manager.current_context = self # tell the context manager about the changed binding

        super(GlutWindow, self).__init__()

        if hide:
            self.hide()
        else:
            self.show()

    def destroy(self):
        _gl.glutDestroyWindow(self._id)
        self._id = 0

    def _bind(self):
        if self._id == 0:
            raise RuntimeError("window has already been destroyed")
        _gl.glutSetWindow(self._id)

    def swap_buffers(self):
        with self:
            _gl.glutSwapBuffers()

    def post_redisplay(self):
        with self:
            _gl.glutPostRedisplay()

    def full_screen(self):
        with self:
            _gl.glutFullScreen()

    def leave_full_screen(self):
        with self:
            _gl.glutLeaveFullScreen()

    def toggle_full_screen(self):
        with self:
            _gl.glutFullScreenToggle()

    def push(self):
        with self:
            _gl.glutPushWindow()

    def pop(self):
        with self:
            _gl.glutPopWindow()

    def show(self):
        with self:
            _gl.glutShowWindow()

    def hide(self):
        with self:
            _gl.glutHideWindow()

    def iconify(self):
        with self:
            _gl.glutIconifyWindow()

    @property
    def cursor(self):
        with self:
            return GlutWindow.cursors[_gl.glutGet(_gl.GLUT_WINDOW_CURSOR)]

    @cursor.setter
    def cursor(self, cursor):
        with self:
            _gl.glutSetCursor(cursor._value)

    @property
    def num_children(self):
        with self:
            return _gl.glutGet(_gl.GLUT_WINDOW_NUM_CHILDREN)

    @property
    def shape(self):
        with self:
            return _gl.glutGet(_gl.GLUT_WINDOW_HEIGHT), _gl.glutGet(_gl.GLUT_WINDOW_WIDTH)

    @shape.setter
    def shape(self, shape):
        height, width = shape
        with self:
            _gl.glutReshapeWindow(width, height)

    @property
    def position(self):
        with self:
            return _gl.glutGet(_gl.GLUT_WINDOW_Y), _gl.glutGet(_gl.GLUT_WINDOW_X)

    @position.setter
    def position(self, position):
        y, x = position
        with self:
            _gl.glutPositionWindow(x, y)

    @property
    def name(self):
        return self._name

    @property
    def window_title(self):
        return self._window_title

    @window_title.setter
    def window_title(self, window_title):
        self._window_title = window_title
        with self:
            _gl.glutSetWindowTitle(window_title.encode("utf8"))

    @property
    def icon_title(self):
        return self._icon_title

    @icon_title.setter
    def icon_title(self, icon_title):
        self._icon_title = icon_title
        with self:
            _gl.glutSetIconTitle(icon_title.encode("utf8"))

    def add_timer(self, msecs, func):
        timer_id = _random.randint(0, 1 << 30)
        while timer_id in self._timer_funcs:
            timer_id = _random.randint(0, 1 << 30)
        self._timer_funcs[timer_id] = func
        _gl.glutTimerFunc(msecs, self._timer_func, timer_id)

    close_callback = _func_property(_gl.glutCloseFunc)
    display_callback = _func_property(_gl.glutDisplayFunc)
    entry_callback = _func_property(_gl.glutEntryFunc)
    idle_callback = _func_property(_gl.glutIdleFunc) # within the callback, current window is not necessarily this window
    joystick_callback = _func_property(_gl.glutJoystickFunc)
    keyboard_callback = _func_property(_gl.glutKeyboardFunc)
    keyboard_up_callback = _func_property(_gl.glutKeyboardUpFunc)
    menu_destroy_callback = _func_property(_gl.glutMenuDestroyFunc)
    menu_state_callback = _func_property(_gl.glutMenuStateFunc)
    menu_status_callback = _func_property(_gl.glutMenuStatusFunc)
    motion_callback = _func_property(_gl.glutMotionFunc)
    mouse_callback = _func_property(_gl.glutMouseFunc)
    mouse_wheel_callback = _func_property(_gl.glutMouseWheelFunc)
    passive_motion_callback = _func_property(_gl.glutPassiveMotionFunc)
    reshape_callback = _func_property(_gl.glutReshapeFunc)
    special_callback = _func_property(_gl.glutSpecialFunc)
    special_up_callback = _func_property(_gl.glutSpecialUpFunc)
    visibility_callback = _func_property(_gl.glutVisibilityFunc)
    window_status_callback = _func_property(_gl.glutWindowStatusFunc)
    wm_close_callback = _func_property(_gl.glutWMCloseFunc)

initialize()

__all__ = [
    "main_loop",
    "main_loop_event",
    "leave_main_loop",
    "get_screen_shape",
    "get_screen_shape_mm",
    "get_elapsed_time",
    "get_shift_state",
    "get_ctrl_state",
    "get_alt_state",
    "GlutWindow",
]




"""Program pipeline class.

@bug: Program pipelines are currently unimplemented.
@todo: Use C{glUseProgramStages}, C{glValidateProgramPipeline} etc.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, BindableObject

class ProgramPipeline(ManagedObject, BindableObject):
    _generate_id = _gl.glGenProgramPipelines
    _delete_id = _gl.glDeleteProgramPipelines
    _db = "program_pipelines"
    _binding = "program_pipeline_binding"

    def __init__(self, context=None):
        super(ProgramPipeline, self).__init__(context=context)

__all__ = ["ProgramPipeline"]




"""Mesh loading methods.

@todo: Implement loading from obj, ply etc.

@author: Stephan Wenger
@date: 2012-03-06
"""
import collections

try:
    import os
    import h5py
    from glitter.convenience.pipeline import Pipeline
    from glitter.convenience.defaultpipeline import get_default_program

    def load_mesh(filename, context=None, **kwargs):
        loader = globals().get("load_%s" % (os.path.splitext(filename)[1][1:].lower()))
        if not loader:
            raise ValueError("no loader for '%s'" % filename)
        return loader(filename, context=context, **kwargs)

    def load_hdf5(filename, color_defaults_to_position=True, context=None, **kwargs):
        with h5py.File(filename, "r") as f:
            kwargs["in_position"] = f["vertices"].value
            if "colors" in f:
                kwargs["in_color"] = f["colors"].value
            elif color_defaults_to_position:
                kwargs["in_color"] = kwargs["in_position"]
            if "indices" in f:
                kwargs["elements"] = f["indices"].value
        return Pipeline(get_default_program(color=("in_color" in kwargs), context=context), use_framebuffer=False, **kwargs)

    load_mesh.supported_formats = tuple(sorted(key[5:] for key, value in list(globals().items())
            if key.startswith("load_") and isinstance(value, collections.Callable) and value is not load_mesh))

    __all__ = ["load_mesh", "load_hdf5"]
except ImportError:
    pass




"""Logging of OpenGL calls.
"""

import re
import collections

class LoggingWrapper(object):
    """Wrap a ctypes function so that function calls are logged.
    """

    def __init__(self, func, logger_func):
        self._func = func
        self._logger_func = logger_func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

        for key in dir(func):
            if not key.startswith("_"):
                setattr(self, key, getattr(func, key))

    def _format_arg(self, x):
        if hasattr(x, "_type_") and hasattr(x, "__len__") and hasattr(x, "__getitem__"):
            return "%s[%s]" % (x._type_.__name__, ", ".join(self._format_arg(a) for a in x))
        elif hasattr(x, "value"):
            return "%s(%r)" % (x.__class__.__name__, x.value)
        elif hasattr(x, "contents"):
            return "&%s" % self._format_arg(x.contents)
        elif hasattr(x, "_name"):
            return x._name
        else:
            return repr(x)

    def _format_args(self, args, kwargs):
        return ", ".join([self._format_arg(x) for x in args] + ["%s=%s" % (k, self._format_arg(v)) for (k, v) in list(kwargs.items())])

    def __call__(self, *args, **kwargs):
        self._logger_func("%s(%s)", self.__name__, self._format_args(args, kwargs))
        return self._func(*args, **kwargs)

    def __str__(self):
        return self.__name__

def add_logger(logger="root", name_re="^(gl|glu|glut|glX)[A-Z].*$", d=None):
    """Add a logger to OpenGL functions.

    All values in C{d} that match C{name_re} have an C{errcheck} attribute will
    be replaced by corresponding L{LoggingWrapper}s that call C{logger} before
    each invocation. By default, C{d} is C{glitter.raw.__dict__}.

    C{logger} may be either a callable, a C{logging.logger} object, the name of
    a registered C{logging.Logger} object, or C{None} to remove the logger.

    @attention: If you add multiple loggers to a function, you will not only
    incur a double performance penalty, but also have to remove them in the
    reverse order; there is no way to remove a specific logger.
    """

    if d is None:
        from glitter import raw
        d = raw.__dict__

    if logger is None:
        for key, value in list(d.items()):
            if re.match(name_re, key) and isinstance(value, LoggingWrapper):
                d[key] = value._func
    else:
        if not isinstance(logger, collections.Callable):
            import logging
            if isinstance(logger, logging.Logger):
                logger = logger.debug
            else:
                logger = logging.getLogger(logger).debug

        for key, value in list(d.items()):
            if re.match(name_re, key) and hasattr(value, "errcheck"):
                d[key] = LoggingWrapper(value, logger)

__all__ = ["LoggingWrapper", "add_logger"]




"""Transform feedback class.

@bug: Transform feedback is currently unimplemented.
@todo: Implement this using C{glBeginTransformFeedback},
C{glPauseTransformFeedback}, C{glResumeTransformFeedback},
C{glEndTransformFeedback}, C{glDrawTransformFeedback},
C{glDrawTransformFeedbackInstanced}, C{glDrawTransformFeedbackStream}, and
C{glDrawTransformFeedbackStreamInstanced}.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, BindableObject

class TransformFeedback(ManagedObject, BindableObject):
    _generate_id = _gl.glGenTransformFeedbacks
    _delete_id = _gl.glDeleteTransformFeedbacks
    _db = "transform_feedbacks"
    _binding = "transform_feedback_binding"
    _target = _gl.GL_TRANSFORM_FEEDBACK

    def __init__(self, context=None):
        super(TransformFeedback, self).__init__(context=context)

__all__ = ["TransformFeedback"]




"""GLX context creation and management.

Depends on the binary glxcontext module.

@todo: Include glxcontext in distribution as an optional module.

@author: Stephan Wenger
@date: 2012-08-28
"""

try:
    from glxcontext import GLXContext as _GLXContext

    from glitter.contexts.context import Context

    class GLXContext(_GLXContext, Context):
        """Offscreen GLX context."""

        def __init__(self, **kwargs):
            _GLXContext.__init__(self, **kwargs)
            Context.__init__(self)

        def _bind(self):
            return _GLXContext.bind(self)

        def bind(self):
            return Context.bind(self)

    __all__ = ["GLXContext"]
except ImportError:
    pass




"""Base class for OpenGL contexts.

This module wraps large parts of per-context state.

@todo: Provide convenience setters for C{glBlendFunc} via C{blend_func=("SRC_ALPHA", "ONE_MINUS_SRC_ALPHA")}.
@todo: Add listing of all bound textures, buffers, etc.
@todo: Do we need indexed variants for C{glEnable}?
@todo: Wrap C{glPatchParameter}.
@todo: Implement multiple viewports using C{glViewportIndexed}, C{glViewportArray}, C{glScissorIndexed}, and C{glScissorArray}.
@todo: Wrap C{glSampleCoverage}, C{glSampleMaski}, C{glMinSampleShading}, C{GL_SAMPLE_COVERAGE_VALUE}, and C{GL_SAMPLE_COVERAGE_INVERT}.
@todo: Wrap C{glReadPixels} and related framebuffer copy functions.
@todo: Add indexed getters for C{GL_TRANSFORM_FEEDBACK_BUFFER_START},
C{GL_TRANSFORM_FEEDBACK_BUFFER_SIZE}, C{GL_UNIFORM_BUFFER_SIZE} and
C{GL_UNIFORM_BUFFER_START}.
@todo: Add getters for C{GL_CLIP_DISTANCEi}, C{GL_DEPTH_CLAMP},
C{GL_FRAMEBUFFER_SRGB}, C{GL_MULTISAMPLE}, C{GL_PRIMITIVE_RESTART}, 
C{GL_TEXTURE_CUBE_MAP_SEAMLESS}, C{GL_SAMPLE_ALPHA_TO_ONE},
C{GL_SAMPLE_ALPHA_TO_COVERAGE}, C{GL_SAMPLE_COVERAGE}, C{GL_SAMPLE_SHADING},
and C{GL_SAMPLE_MASK}.
@todo: Implement stencil buffers using C{GL_STENCIL_BACK_FUNC},
C{GL_STENCIL_FUNC}, C{GL_STENCIL_BACK_REF}, C{GL_STENCIL_REF},
C{GL_STENCIL_BACK_VALUE_MASK}, C{GL_STENCIL_BACK_WRITEMASK},
C{GL_STENCIL_VALUE_MASK}, C{GL_STENCIL_WRITEMASK}, C{GL_STENCIL_BACK_FAIL},
C{GL_STENCIL_BACK_PASS_DEPTH_FAIL}, C{GL_STENCIL_BACK_PASS_DEPTH_PASS},
C{GL_STENCIL_FAIL}, C{GL_STENCIL_PASS_DEPTH_FAIL},
C{GL_STENCIL_PASS_DEPTH_PASS}, etc.

@author: Stephan Wenger
@date: 2012-02-29
"""

from weakref import WeakValueDictionary

import glitter.raw as _gl
from glitter.utils import blend_functions, blend_equations, depth_functions, draw_buffers, hints, provoking_vertices, logic_op_modes, provoke_modes, color_read_formats, color_read_types, read_buffers, cull_face_modes, front_face_modes, polygon_modes, InstanceDescriptorMixin, State, StateMixin
from glitter.contexts.contextmanager import context_manager
from glitter.contexts.proxies import BooleanProxy, FloatProxy, IntegerProxy, Integer64Proxy, EnableDisableProxy, EnumProxy, StringProxy, HintProxy, BindingProxy
from glitter.contexts.textures import TextureUnitList
from glitter.contexts.drawbuffers import DrawBufferList, ColorWritemaskList
from glitter.contexts.multiproxies import BlendFuncProxy, BlendEquationProxy, PolygonOffsetProxy

class GLObjectLibrary(object):
    def __init__(self, _context):
        self._context = _context
        self._objects = WeakValueDictionary()

    def __getitem__(self, key):
        return self._objects[key]

    def __str__(self):
        return str(dict(self._objects))

    def __repr__(self):
        return str(self)

class Context(InstanceDescriptorMixin, StateMixin):
    _frozen = False

    def __init__(self):
        self.texture_units = TextureUnitList(self)
        self.color_writemasks = ColorWritemaskList(self)
        self.draw_buffers = DrawBufferList(self)
        self.buffers = GLObjectLibrary(self)
        self.framebuffers = GLObjectLibrary(self)
        self.program_pipelines = GLObjectLibrary(self)
        self.queries = GLObjectLibrary(self)
        self.renderbuffers = GLObjectLibrary(self)
        self.samplers = GLObjectLibrary(self)
        self.shader_programs = GLObjectLibrary(self)
        self.shaders = GLObjectLibrary(self)
        self.textures = GLObjectLibrary(self)
        self.transform_feedbacks = GLObjectLibrary(self)
        self.vertex_arrays = GLObjectLibrary(self)

        super(Context, self).__init__()

        self._frozen = True

    def __setattr__(self, name, value):
        if self._frozen and not name.startswith("_") and not hasattr(self, name):
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))
        return super(Context, self).__setattr__(name, value)

    def _bind(self):
        """Make the context active.

        This should be overwritten appropriately by window system dependent subclasses.
        """

        pass

    def bind(self):
        old_binding = context_manager.current_context
        context_manager.current_context = self
        return old_binding

    def __enter__(self):
        context_manager.push(self)
        return self

    def __exit__(self, type, value, traceback):
        context_manager.pop()

    #{ Enums
    blend_functions = blend_functions
    blend_equations = blend_equations
    depth_functions = depth_functions
    draw_buffers = draw_buffers
    hints = hints
    provoking_vertices = provoking_vertices
    logic_op_modes = logic_op_modes
    provoke_modes = provoke_modes
    color_read_formats = color_read_formats
    color_read_types = color_read_types
    read_buffers = read_buffers
    cull_face_modes = cull_face_modes
    front_face_modes = front_face_modes
    polygon_modes = polygon_modes

    #{ Buffer bindings
    array_buffer_binding                 = BindingProxy(_gl.glBindBuffer,          [_gl.GL_ARRAY_BUFFER                ])
    atomic_counter_buffer_binding        = BindingProxy(_gl.glBindBuffer,          [_gl.GL_ATOMIC_COUNTER_BUFFER       ])
    copy_read_buffer_binding             = BindingProxy(_gl.glBindBuffer,          [_gl.GL_COPY_READ_BUFFER            ])
    copy_write_buffer_binding            = BindingProxy(_gl.glBindBuffer,          [_gl.GL_COPY_WRITE_BUFFER           ])
    draw_indirect_buffer_binding         = BindingProxy(_gl.glBindBuffer,          [_gl.GL_DRAW_INDIRECT_BUFFER        ])
    element_array_buffer_binding         = BindingProxy(_gl.glBindBuffer,          [_gl.GL_ELEMENT_ARRAY_BUFFER        ])
    pixel_pack_buffer_binding            = BindingProxy(_gl.glBindBuffer,          [_gl.GL_PIXEL_PACK_BUFFER           ])
    pixel_unpack_buffer_binding          = BindingProxy(_gl.glBindBuffer,          [_gl.GL_PIXEL_UNPACK_BUFFER         ])
    texture_buffer_binding               = BindingProxy(_gl.glBindBuffer,          [_gl.GL_TEXTURE_BUFFER              ])
    transform_feedback_buffer_binding    = BindingProxy(_gl.glBindBuffer,          [_gl.GL_TRANSFORM_FEEDBACK_BUFFER   ])
    uniform_buffer_binding               = BindingProxy(_gl.glBindBuffer,          [_gl.GL_UNIFORM_BUFFER              ])

    #{ Miscellaneous bindings
    program_pipeline_binding             = BindingProxy(_gl.glBindProgramPipeline, [                                   ])
    renderbuffer_binding                 = BindingProxy(_gl.glBindRenderbuffer,    [_gl.GL_RENDERBUFFER                ])
    vertex_array_binding                 = BindingProxy(_gl.glBindVertexArray,     [                                   ])
    draw_framebuffer_binding             = BindingProxy(_gl.glBindFramebuffer,     [_gl.GL_DRAW_FRAMEBUFFER            ])
    read_framebuffer_binding             = BindingProxy(_gl.glBindFramebuffer,     [_gl.GL_READ_FRAMEBUFFER            ])
    current_program                      = BindingProxy(_gl.glUseProgram,          [                                   ])
    active_texture                       = BindingProxy(_gl.glActiveTexture,       [                                   ])

    #{ Properties with separate getters but joint setters
    blend_dst_alpha = BlendFuncProxy(_gl.GL_BLEND_DST_ALPHA)
    blend_dst_rgb = BlendFuncProxy(_gl.GL_BLEND_DST_RGB)
    blend_src_alpha = BlendFuncProxy(_gl.GL_BLEND_SRC_ALPHA)
    blend_src_rgb = BlendFuncProxy(_gl.GL_BLEND_SRC_RGB)
    blend_equation_alpha = BlendEquationProxy(_gl.GL_BLEND_EQUATION_ALPHA)
    blend_equation_rgb = BlendEquationProxy(_gl.GL_BLEND_EQUATION_RGB)
    polygon_offset_factor = PolygonOffsetProxy(_gl.GL_POLYGON_OFFSET_FACTOR)
    polygon_offset_units = PolygonOffsetProxy(_gl.GL_POLYGON_OFFSET_UNITS)

    #{ Miscellaneous enumerations
    depth_func = EnumProxy(depth_functions, _gl.GL_DEPTH_FUNC, _gl.glDepthFunc)
    draw_buffer = EnumProxy(draw_buffers, _gl.GL_DRAW_BUFFER, _gl.glDrawBuffer)
    implementation_color_read_format = EnumProxy(color_read_formats, _gl.GL_IMPLEMENTATION_COLOR_READ_FORMAT)
    implementation_color_read_type = EnumProxy(color_read_types, _gl.GL_IMPLEMENTATION_COLOR_READ_TYPE)
    layer_provoking_vertex = EnumProxy(provoking_vertices, _gl.GL_LAYER_PROVOKING_VERTEX)
    logic_op_mode = EnumProxy(logic_op_modes, _gl.GL_LOGIC_OP_MODE, _gl.glLogicOp)
    provoking_vertex = EnumProxy(provoke_modes, _gl.GL_PROVOKING_VERTEX, _gl.glProvokingVertex)
    read_buffer = EnumProxy(read_buffers, _gl.GL_READ_BUFFER, _gl.glReadBuffer)
    viewport_index_provoking_vertex = EnumProxy(provoking_vertices, _gl.GL_VIEWPORT_INDEX_PROVOKING_VERTEX)
    cull_face_mode = EnumProxy(cull_face_modes, _gl.GL_CULL_FACE_MODE, _gl.glCullFace)
    front_face = EnumProxy(front_face_modes, _gl.GL_FRONT_FACE, _gl.glFrontFace)
    polygon_mode = EnumProxy(polygon_modes, _gl.GL_POLYGON_MODE, _gl.glPolygonMode, [_gl.GL_FRONT_AND_BACK])

    #{ Hints
    fragment_shader_derivative_hint = HintProxy(_gl.GL_FRAGMENT_SHADER_DERIVATIVE_HINT)
    line_smooth_hint = HintProxy(_gl.GL_LINE_SMOOTH_HINT)
    polygon_smooth_hint = HintProxy(_gl.GL_POLYGON_SMOOTH_HINT)
    texture_compression_hint = HintProxy(_gl.GL_TEXTURE_COMPRESSION_HINT)

    #{ Enabling and disabling
    blend = EnableDisableProxy(_gl.GL_BLEND)
    color_logic_op = EnableDisableProxy(_gl.GL_COLOR_LOGIC_OP)
    cull_face = EnableDisableProxy(_gl.GL_CULL_FACE)
    depth_test = EnableDisableProxy(_gl.GL_DEPTH_TEST)
    dither = EnableDisableProxy(_gl.GL_DITHER)
    line_smooth = EnableDisableProxy(_gl.GL_LINE_SMOOTH)
    polygon_offset_fill = EnableDisableProxy(_gl.GL_POLYGON_OFFSET_FILL)
    polygon_offset_line = EnableDisableProxy(_gl.GL_POLYGON_OFFSET_LINE)
    polygon_offset_point = EnableDisableProxy(_gl.GL_POLYGON_OFFSET_POINT)
    polygon_smooth = EnableDisableProxy(_gl.GL_POLYGON_SMOOTH)
    scissor_test = EnableDisableProxy(_gl.GL_SCISSOR_TEST)
    stencil_test = EnableDisableProxy(_gl.GL_STENCIL_TEST)
    vertex_program_point_size = EnableDisableProxy(_gl.GL_VERTEX_PROGRAM_POINT_SIZE)

    #{ Boolean values
    color_writemask = BooleanProxy([_gl.GL_COLOR_WRITEMASK], _gl.glColorMask, shape=4)
    depth_writemask = BooleanProxy([_gl.GL_DEPTH_WRITEMASK], _gl.glDepthMask)
    doublebuffer = BooleanProxy([_gl.GL_DOUBLEBUFFER])
    pack_lsb_first = BooleanProxy([_gl.GL_PACK_LSB_FIRST], _gl.glPixelStorei, [_gl.GL_PACK_LSB_FIRST])
    pack_swap_bytes = BooleanProxy([_gl.GL_PACK_SWAP_BYTES], _gl.glPixelStorei, [_gl.GL_PACK_SWAP_BYTES])
    shader_compiler = BooleanProxy([_gl.GL_SHADER_COMPILER])
    stereo = BooleanProxy([_gl.GL_STEREO])
    unpack_lsb_first = BooleanProxy([_gl.GL_UNPACK_LSB_FIRST], _gl.glPixelStorei, [_gl.GL_UNPACK_LSB_FIRST])
    unpack_swap_bytes = BooleanProxy([_gl.GL_UNPACK_SWAP_BYTES], _gl.glPixelStorei, [_gl.GL_UNPACK_SWAP_BYTES])
    
    #{ Float values
    aliased_line_width_range = FloatProxy([_gl.GL_ALIASED_LINE_WIDTH_RANGE], shape=2)
    blend_color = FloatProxy([_gl.GL_BLEND_COLOR], _gl.glBlendColor, shape=4)
    color_clear_value = FloatProxy([_gl.GL_COLOR_CLEAR_VALUE], _gl.glClearColor, shape=4)
    depth_clear_value = FloatProxy([_gl.GL_DEPTH_CLEAR_VALUE], _gl.glClearDepth)
    depth_range = FloatProxy([_gl.GL_DEPTH_RANGE], _gl.glDepthRange, shape=2)
    line_width = FloatProxy([_gl.GL_LINE_WIDTH], _gl.glLineWidth)
    line_width_granularity = FloatProxy([_gl.GL_LINE_WIDTH_GRANULARITY])
    line_width_range = FloatProxy([_gl.GL_LINE_WIDTH_RANGE], shape=2)
    point_fade_threshold_size = FloatProxy([_gl.GL_POINT_FADE_THRESHOLD_SIZE], _gl.glPointParameterf, [_gl.GL_POINT_FADE_THRESHOLD_SIZE])
    point_size = FloatProxy([_gl.GL_POINT_SIZE], _gl.glPointSize)
    point_size_granularity = FloatProxy([_gl.GL_POINT_SIZE_GRANULARITY])
    point_size_range = FloatProxy([_gl.GL_POINT_SIZE_RANGE], shape=2)
    smooth_line_width_range = FloatProxy([_gl.GL_SMOOTH_LINE_WIDTH_RANGE], shape=2)
    smooth_line_width_granularity = FloatProxy([_gl.GL_SMOOTH_LINE_WIDTH_GRANULARITY])

    #{ Integer values
    major_version = IntegerProxy([_gl.GL_MAJOR_VERSION])
    max_3d_texture_size = IntegerProxy([_gl.GL_MAX_3D_TEXTURE_SIZE])
    max_array_texture_layers = IntegerProxy([_gl.GL_MAX_3D_TEXTURE_SIZE])
    max_clip_distances = IntegerProxy([_gl.GL_MAX_CLIP_DISTANCES])
    max_color_texture_samples = IntegerProxy([_gl.GL_MAX_COLOR_TEXTURE_SAMPLES])
    max_color_attachments = IntegerProxy([_gl.GL_MAX_COLOR_ATTACHMENTS])
    max_combined_fragment_uniform_components = IntegerProxy([_gl.GL_MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS])
    max_combined_geometry_uniform_components = IntegerProxy([_gl.GL_MAX_COMBINED_GEOMETRY_UNIFORM_COMPONENTS])
    max_combined_texture_image_units = IntegerProxy([_gl.GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS])
    max_combined_uniform_blocks = IntegerProxy([_gl.GL_MAX_COMBINED_UNIFORM_BLOCKS])
    max_combined_vertex_uniform_components = IntegerProxy([_gl.GL_MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS])
    max_cube_map_texture_size = IntegerProxy([_gl.GL_MAX_CUBE_MAP_TEXTURE_SIZE])
    max_depth_texture_samples = IntegerProxy([_gl.GL_MAX_DEPTH_TEXTURE_SAMPLES])
    max_draw_buffers = IntegerProxy([_gl.GL_MAX_DRAW_BUFFERS])
    max_elements_indices = IntegerProxy([_gl.GL_MAX_ELEMENTS_INDICES])
    max_elements_vertices = IntegerProxy([_gl.GL_MAX_ELEMENTS_VERTICES])
    max_fragment_input_components = IntegerProxy([_gl.GL_MAX_FRAGMENT_INPUT_COMPONENTS])
    max_fragment_uniform_components = IntegerProxy([_gl.GL_MAX_FRAGMENT_UNIFORM_COMPONENTS])
    max_fragment_uniform_vectors = IntegerProxy([_gl.GL_MAX_FRAGMENT_UNIFORM_VECTORS])
    max_fragment_uniform_blocks = IntegerProxy([_gl.GL_MAX_FRAGMENT_UNIFORM_BLOCKS])
    max_geometry_input_components = IntegerProxy([_gl.GL_MAX_GEOMETRY_INPUT_COMPONENTS])
    max_geometry_output_components = IntegerProxy([_gl.GL_MAX_GEOMETRY_OUTPUT_COMPONENTS])
    max_geometry_texture_image_units = IntegerProxy([_gl.GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS])
    max_geometry_uniform_blocks = IntegerProxy([_gl.GL_MAX_GEOMETRY_UNIFORM_BLOCKS])
    max_geometry_uniform_components = IntegerProxy([_gl.GL_MAX_GEOMETRY_UNIFORM_COMPONENTS])
    max_integer_samples = IntegerProxy([_gl.GL_MAX_INTEGER_SAMPLES])
    max_program_texel_offset = IntegerProxy([_gl.GL_MAX_PROGRAM_TEXEL_OFFSET])
    min_program_texel_offset = IntegerProxy([_gl.GL_MIN_PROGRAM_TEXEL_OFFSET])
    max_rectangle_texture_size = IntegerProxy([_gl.GL_MAX_RECTANGLE_TEXTURE_SIZE])
    max_renderbuffer_size = IntegerProxy([_gl.GL_MAX_RENDERBUFFER_SIZE])
    max_sample_mask_words = IntegerProxy([_gl.GL_MAX_SAMPLE_MASK_WORDS])
    max_server_wait_timeout = IntegerProxy([_gl.GL_MAX_SERVER_WAIT_TIMEOUT])
    max_texture_buffer_size = IntegerProxy([_gl.GL_MAX_TEXTURE_BUFFER_SIZE])
    max_texture_image_units = IntegerProxy([_gl.GL_MAX_TEXTURE_IMAGE_UNITS])
    max_texture_lod_bias = IntegerProxy([_gl.GL_MAX_TEXTURE_LOD_BIAS])
    max_texture_size = IntegerProxy([_gl.GL_MAX_TEXTURE_SIZE])
    max_uniform_buffer_bindings = IntegerProxy([_gl.GL_MAX_UNIFORM_BUFFER_BINDINGS])
    max_uniform_block_size = IntegerProxy([_gl.GL_MAX_UNIFORM_BLOCK_SIZE])
    max_varying_vectors = IntegerProxy([_gl.GL_MAX_VARYING_VECTORS])
    max_vertex_attribs = IntegerProxy([_gl.GL_MAX_VERTEX_ATTRIBS])
    max_vertex_texture_image_units = IntegerProxy([_gl.GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS])
    max_vertex_uniform_components = IntegerProxy([_gl.GL_MAX_VERTEX_UNIFORM_COMPONENTS])
    max_vertex_uniform_vectors = IntegerProxy([_gl.GL_MAX_VERTEX_UNIFORM_VECTORS])
    max_vertex_output_components = IntegerProxy([_gl.GL_MAX_VERTEX_OUTPUT_COMPONENTS])
    max_vertex_uniform_blocks = IntegerProxy([_gl.GL_MAX_VERTEX_UNIFORM_BLOCKS])
    max_viewport_dims = IntegerProxy([_gl.GL_MAX_VIEWPORT_DIMS])
    max_viewports = IntegerProxy([_gl.GL_MAX_VIEWPORTS])
    minor_version = IntegerProxy([_gl.GL_MINOR_VERSION])
    pack_alignment = IntegerProxy([_gl.GL_PACK_ALIGNMENT])
    pack_image_height = IntegerProxy([_gl.GL_PACK_IMAGE_HEIGHT])
    pack_row_length = IntegerProxy([_gl.GL_PACK_ROW_LENGTH], _gl.glPixelStorei, [_gl.GL_PACK_ROW_LENGTH])
    pack_skip_images = IntegerProxy([_gl.GL_PACK_SKIP_IMAGES], _gl.glPixelStorei, [_gl.GL_PACK_SKIP_IMAGES])
    pack_skip_pixels = IntegerProxy([_gl.GL_PACK_SKIP_PIXELS], _gl.glPixelStorei, [_gl.GL_PACK_SKIP_PIXELS])
    pack_skip_rows = IntegerProxy([_gl.GL_PACK_SKIP_ROWS], _gl.glPixelStorei, [_gl.GL_PACK_SKIP_ROWS])
    primitive_restart_index = IntegerProxy([_gl.GL_PRIMITIVE_RESTART_INDEX], _gl.glPrimitiveRestartIndex)
    sample_buffers = IntegerProxy([_gl.GL_SAMPLE_BUFFERS])
    samples = IntegerProxy([_gl.GL_SAMPLES])
    scissor_box = IntegerProxy([_gl.GL_SCISSOR_BOX], _gl.glScissor, shape=4)
    stencil_clear_value = IntegerProxy([_gl.GL_STENCIL_CLEAR_VALUE], _gl.glClearStencil)
    subpixel_bits = IntegerProxy([_gl.GL_SUBPIXEL_BITS])
    uniform_buffer_offset_alignment = IntegerProxy([_gl.GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT])
    unpack_alignment = IntegerProxy([_gl.GL_UNPACK_ALIGNMENT], _gl.glPixelStorei, [_gl.GL_UNPACK_ALIGNMENT])
    unpack_image_height = IntegerProxy([_gl.GL_UNPACK_IMAGE_HEIGHT], _gl.glPixelStorei, [_gl.GL_UNPACK_IMAGE_HEIGHT])
    unpack_row_length = IntegerProxy([_gl.GL_UNPACK_ROW_LENGTH], _gl.glPixelStorei, [_gl.GL_UNPACK_ROW_LENGTH])
    unpack_skip_images = IntegerProxy([_gl.GL_UNPACK_SKIP_IMAGES], _gl.glPixelStorei, [_gl.GL_UNPACK_SKIP_IMAGES])
    unpack_skip_pixels = IntegerProxy([_gl.GL_UNPACK_SKIP_PIXELS], _gl.glPixelStorei, [_gl.GL_UNPACK_SKIP_PIXELS])
    unpack_skip_rows = IntegerProxy([_gl.GL_UNPACK_SKIP_ROWS], _gl.glPixelStorei, [_gl.GL_UNPACK_SKIP_ROWS])
    viewport = IntegerProxy([_gl.GL_VIEWPORT], _gl.glViewport, shape=4) # XXX viewport dimensions are (x, y, width, height) as in OpenGL, not (height, width) as in numpy!
    viewport_bounds_range = IntegerProxy([_gl.GL_VIEWPORT_BOUNDS_RANGE], shape=2)
    viewport_subpixel_bits = IntegerProxy([_gl.GL_VIEWPORT_SUBPIXEL_BITS])

    #{ 64-bit integer values
    timestamp = Integer64Proxy([_gl.GL_TIMESTAMP])

    #{ Strings
    vendor = StringProxy(_gl.GL_VENDOR)
    renderer = StringProxy(_gl.GL_RENDERER)
    version = StringProxy(_gl.GL_VERSION)
    shading_language_version = StringProxy(_gl.GL_SHADING_LANGUAGE_VERSION)
    extensions = StringProxy(_gl.GL_EXTENSIONS, _gl.GL_NUM_EXTENSIONS)

    #{ Global actions

    def _perform_gl_clear(self, color=None, depth=None, stencil=None):
        """Call C{glClear} with appropriate arguments.

        @param color: Whether to clear the color buffer, and optionally, to which value.
        @type color: C{bool} or C{numpy.ndarray}.
        @param depth: Whether to clear the depth buffer, and optionally, to which value.
        @type depth: C{bool} or C{numpy.ndarray}.
        @param stencil: Whether to clear the stencil buffer, and optionally, to which value.
        @type stencil: C{bool} or C{numpy.ndarray}.

        If no parameters are given, color, depth and stencil are cleared with
        the current clear values.

        @todo: Wrap the C{glClear} call in C{with State()} to set and reset
        L{color_clear_value}, C{depth_clear_value} and C{stencil_clear_value}
        if these are given as C{numpy} arrays in C{color}, C{depth}, and
        C{stencil}, respectively.
        """

        if color is None and depth is None and stencil is None:
            color = depth = stencil = True
        _gl.glClear(
                (_gl.GL_COLOR_BUFFER_BIT if color else 0) |
                (_gl.GL_DEPTH_BUFFER_BIT if depth else 0) |
                (_gl.GL_STENCIL_BUFFER_BIT if stencil else 0)
                )

    def clear(self, color=None, depth=None, stencil=None):
        """Clear the default framebuffer.

        @param color: Whether to clear the color buffer, and optionally, to which value.
        @type color: C{bool} or C{numpy.ndarray}.
        @param depth: Whether to clear the depth buffer, and optionally, to which value.
        @type depth: C{bool} or C{numpy.ndarray}.
        @param stencil: Whether to clear the stencil buffer, and optionally, to which value.
        @type stencil: C{bool} or C{numpy.ndarray}.

        If no parameters are given, color, depth and stencil are cleared with
        the current clear values.
        """

        with self:
            with State(self, draw_framebuffer_binding=None):
                self._perform_gl_clear(color, depth, stencil)

    def finish(self):
        """Block until all GL execution is complete.

        C{finish} does not return until the effects of all previously called GL
        commands are complete. Such effects include all changes to GL state,
        all changes to connection state, and all changes to the frame buffer
        contents.
        """

        with self:
            _gl.glFinish()

    def flush(self):
        """Force execution of GL commands in finite time.

        Different GL implementations buffer commands in several different
        locations, including network buffers and the graphics accelerator
        itself. C{flush} empties all of these buffers, causing all issued
        commands to be executed as quickly as they are accepted by the actual
        rendering engine. Though this execution may not be completed in any
        particular time period, it does complete in finite time.

        Because any GL program might be executed over a network, or on an
        accelerator that buffers commands, all programs should call C{flush}
        whenever they count on having all of their previously issued commands
        completed. For example, call C{flush} before waiting for user input
        that depends on the generated image.

        C{flush} can return at any time. It does not wait until the execution
        of all previously issued GL commands is complete.
        """

        with self:
            _gl.glFlush()

    def release_shader_compiler(self):
        with self:
            _gl.glReleaseShaderCompiler()

    #}

__all__ = ["Context"]




"""Memory barrier function.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl

def memory_barrier(
    vertex_attrib_array=False, element_array=False, uniform=False,
    texture_fetch=False, shader_image_access=False, command=False,
    pixel_buffer=False, texture_update=False, buffer_update=False,
    framebuffer=False, transform_feedback=False, atomic_counter=False):
    bits = ((_gl.GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT if vertex_attrib_array else 0) |
            (_gl.GL_ELEMENT_ARRAY_BARRIER_BIT if element_array else 0) |
            (_gl.GL_UNIFORM_BARRIER_BIT if uniform else 0) |
            (_gl.GL_TEXTURE_FETCH_BARRIER_BIT if texture_fetch else 0) |
            (_gl.GL_SHADER_IMAGE_ACCESS_BARRIER_BIT if shader_image_access else 0) |
            (_gl.GL_COMMAND_BARRIER_BIT if command else 0) |
            (_gl.GL_PIXEL_BUFFER_BARRIER_BIT if pixel_buffer else 0) |
            (_gl.GL_TEXTURE_UPDATE_BARRIER_BIT if texture_update else 0) |
            (_gl.GL_BUFFER_UPDATE_BARRIER_BIT if buffer_update else 0) |
            (_gl.GL_FRAMEBUFFER_BARRIER_BIT if framebuffer else 0) |
            (_gl.GL_TRANSFORM_FEEDBACK_BARRIER_BIT if transform_feedback else 0) |
            (_gl.GL_ATOMIC_COUNTER_BARRIER_BIT if atomic_counter else 0))
    if bits == 0:
        bits = _gl.GL_ALL_BARRIER_BITS
    _gl.glMemoryBarrier(bits)




"""Datatypes and helper functions.

@author: Stephan Wenger
@date: 2012-02-29
"""

import numpy as _np

import glitter.raw as _gl

class Datatype(object):
    _nptype_db = {}
    _gltype_db = {}
    _db = {}

    def __init__(self, integer, signed, nptype, _gltype=None, charcode=None):
        self._integer = bool(integer)
        self._signed = bool(signed)
        self._nptype = nptype
        self._gltype = _gltype
        self._charcode = charcode
        self._nbytes = nptype().itemsize

        Datatype._nptype_db[self._nptype] = self
        Datatype._gltype_db[self._gltype] = self
        Datatype._db[self._integer, self._signed, self._nbytes] = self

    @classmethod
    def from_numpy(cls, nptype):
        try:
            return cls._nptype_db[nptype]
        except KeyError:
            return cls._nptype_db[nptype.type]

    @classmethod
    def _from_gl(cls, _gltype):
        return cls._gltype_db[_gltype]

    def as_numpy(self):
        return self._nptype

    def _as_gl(self):
        return self._gltype

    def as_signed(self):
        return Datatype._db[self.is_integer(), True, self.nbytes]

    def as_unsigned(self):
        return Datatype._db[self.is_integer(), False, self.nbytes]

    def as_nbytes(self, nbytes):
        return Datatype._db[self.is_integer(), self.is_signed(), nbytes]

    def is_signed(self):
        return self._signed

    def is_unsigned(self):
        return not self._signed

    def is_integer(self):
        return self._integer

    def is_float(self):
        return not self._integer

    def is_boolean(self):
        return self == bool8

    @property
    def charcode(self):
        return self._charcode

    @property
    def nbytes(self):
        return self._nbytes

    def __str__(self):
        return "%s%s%d" % ("u" if self.is_unsigned() else "", "int" if self.is_integer() else "float", 8 * self.nbytes)

    def __repr__(self):
        return str(self)

    def coerced(self, force_integer=False, force_unsigned=False, force_float=False, force_gl=True):
        """Find a datatype with the specified properties.

        The returned datatype will match `self` as closely as possible.
        If `force_integer` is `True`, it will be an integer datatype.
        If `force_unsigned` is `True`, it will be an unsigned datatype.
        If `force_float` is `True`, it will be a floating point datatype.
        If `force_gl` is `True`, it will have an OpenGL equivalent.

        If conversion of `self` to the resulting datatype would likely result in a
        severe loss of precision or overflows, an error may be raised.
        """

        if force_integer and not self.is_integer():
            raise TypeError("no matching datatype")
        if force_float and not self.is_float():
            raise TypeError("no matching datatype")

        dtype = self
        if force_unsigned and not dtype.is_unsigned():
            dtype = dtype.as_unsigned()
        while force_gl and dtype._as_gl() is None:
            dtype = dtype.as_nbytes(dtype.nbytes / 2)
        return dtype

bool8 = Datatype(integer=True, signed=True, nptype=_np.bool8, _gltype=_gl.GL_BOOL, charcode="b")
uint8 = Datatype(integer=True, signed=False, nptype=_np.uint8, _gltype=_gl.GL_UNSIGNED_BYTE, charcode="ub")
uint16 = Datatype(integer=True, signed=False, nptype=_np.uint16, _gltype=_gl.GL_UNSIGNED_SHORT)
uint32 = Datatype(integer=True, signed=False, nptype=_np.uint32, _gltype=_gl.GL_UNSIGNED_INT, charcode="ui")
uint64 = Datatype(integer=True, signed=False, nptype=_np.uint64, charcode="ui64")
int8 = Datatype(integer=True, signed=True, nptype=_np.int8, _gltype=_gl.GL_BYTE)
int16 = Datatype(integer=True, signed=True, nptype=_np.int16, _gltype=_gl.GL_SHORT)
int32 = Datatype(integer=True, signed=True, nptype=_np.int32, _gltype=_gl.GL_INT, charcode="i")
int64 = Datatype(integer=True, signed=True, nptype=_np.int64, charcode="i64")
float32 = Datatype(integer=False, signed=True, nptype=_np.float32, _gltype=_gl.GL_FLOAT, charcode="f")
float64 = Datatype(integer=False, signed=True, nptype=_np.float64, charcode="d") # GL_DOUBLE exists but is not supported by, e.g., textures

def coerce_array(data, dtype=None, force_integer=False, force_unsigned=False, force_float=False, force_gl=True):
    """Convert `data` to a contiguous array with the specified properties.

    The datatype of the array will match `dtype` as closely as possible.
    If `force_integer` is `True`, it will be an integer datatype.
    If `force_unsigned` is `True`, it will be an unsigned datatype.
    If `force_float` is `True`, it will be a floating point datatype.
    If `force_gl` is `True`, it will have an OpenGL equivalent.

    If conversion to the resulting datatype would likely result in a severe
    loss of precision or overflows, an error may be raised.
    """

    if dtype is None:
        if hasattr(data, "dtype"):
            dtype = Datatype.from_numpy(data.dtype)
        else:
            if force_integer:
                dtype = int32
            else:
                dtype = float32
    dtype = dtype.coerced(force_integer, force_unsigned, force_float, force_gl)

    return _np.ascontiguousarray(data, dtype.as_numpy())

__all__ = [
    "Datatype",
    "coerce_array",
    "bool8",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "int8",
    "int16",
    "int32",
    "int64",
    "float32",
    "float64",
]




#!/usr/bin/env python

#! This file is a literate Python program. You can compile the documentation
#! using mylit (http://pypi.python.org/pypi/mylit/).
## title = "glitter Example: Simple"
## stylesheet = "pygments_style.css"

# <h1><i>glitter</i> Example: Simple</h1>

# <h2>Summary</h2>

# This program will open a GLUT window and render a colored, rotating quad.

# <img src="simple.png">

# <h2>Front matter</h2>

# <h3>Module docstring</h3>

# The module docstring is used as a description of this example in the
# generated documentation:
"""Simple example displaying a rotating quad.

@author: Stephan Wenger
@date: 2012-02-29
"""

# <h3>Imports</h3>

# Our scene is going to rotate. The rotating modelview matrix is computed using
# the sine and cosine functions from the math module:
from math import cos, sin

# We can usually import classes and functions contained in <i>glitter</i>
# submodules directly from glitter:
from glitter import VertexArray, get_default_program

# Modules with external dependencies other than numpy, such as platform
# dependent parts like methods for the generation of an OpenGL context,
# however, have to be imported from their respective submodules:
from glitter.contexts.glut import GlutWindow, main_loop, get_elapsed_time

# <h2>Vertex arrays</h2>

# The geometry is specified as arrays (or nested lists) of vertices, colors,
# and indices into the vertex and color arrays. Lists are automatically
# converted to appropriate numpy arrays. The shape of these arrays describes
# how the arrays are to be drawn. When an index array is used, as in this
# example, the vertex and color arrays are two-dimensional: the first dimension
# is the number of vertices, the second is the number of values for each vertex
# (e.g. the red, green, blue, and alpha channels for a color array). The shape
# of the index array is also two-dimensional: the first dimension is the number
# of primitives to be drawn, the second is the number of vertices in each
# primitive. Here, we draw four triangles forming a quad.
vertices = ((0, 0, 0), (-1, 1, 0), (1, 1, 0), (1, -1, 0), (-1, -1, 0))
colors = ((1, 1, 1), (0, 1, 0), (0, 0, 1), (0, 1, 1), (1, 0, 0))
indices = ((0, 1, 2), (0, 2, 3), (0, 3, 4), (0, 4, 1))

# <h2>Main class</h2>

# We wrap all the OpenGL interaction in a class. The class will contain an
# <code>__init__()</code> method to set up all OpenGL objects, any required
# callback methods, as well as a <code>run()</code> method to trigger execution
# of the GLUT main loop.
class SimpleExample(object):
    # <h3>Initialization</h3>

    # When a <code>SimpleExample</code> instance is created, we need to
    # initialize a few OpenGL objects.
    def __init__(self):
        # First, we create a window; this also creates an OpenGL context.
        self.window = GlutWindow(double=True, multisample=True)

        # Then, we set the GLUT display callback function which will be defined
        # later.
        self.window.display_callback = self.display

        # In the OpenGL core profile, there is no such thing as a "standard pipeline"
        # any more. We use the minimalistic <code>defaultpipeline</code> from the
        # <code>glitter.convenience</code> module to create a shader program instead:
        self.shader = get_default_program()

        # Here, we create a vertex array that contains buffers for two vertex array
        # input variables as well as an index array:
        self.vao = VertexArray(vertices, colors, elements=indices)

    # <h3>Callback functions</h3>

    # <h4>Display function</h4>

    # Here we define the display function. It will be called by GLUT whenever the
    # screen has to be redrawn.
    def display(self):
        # First we clear the default framebuffer:
        self.window.clear()
        
        # To draw the vertex array, we use:
        self.vao.draw()

        # After all rendering commands have been issued, we swap the back buffer to
        # the front, making the rendered image visible all at once:
        self.window.swap_buffers()

    # <h4>Timer function</h4>

    # The animation is controlled by a GLUT timer. The timer callback changes the
    # modelview matrix, schedules the next timer event, and causes a screen redraw:
    def timer(self):
        # We first get the elapsed time from GLUT using <code>get_elapsed_time()</code>:
        phi = get_elapsed_time()

        # We then set the <code>modelview_matrix</code> uniform variable of the
        # shader created in the initialization section simply by setting an
        # attribute:
        self.shader.modelview_matrix = ((cos(phi), sin(phi), 0, 0), (-sin(phi), cos(phi), 0, 0), (0, 0, 1, 0), (0, 0, 0, 2))

        # The following line schedules the next timer event to execute after ten milliseconds.
        self.window.add_timer(10, self.timer)

        # Finally, we tell GLUT to redraw the screen.
        self.window.post_redisplay()

    # <h3>Running</h3>
    
    # We will call the <code>run()</code> method later to run the OpenGL code.
    def run(self):
        # To start the animation, we call the timer once; all subsequent timer
        # calls will be scheduled by the timer function itself.
        self.timer()

        # The default program is bound by using a <code>with</code> statement:
        with self.shader:
            # With the shader bound, we enter the GLUT main loop.
            main_loop()

        # When the main loop exits, control is handed back to the script.

# <h2>Main section</h2>

# Finally, if this program is being run from the command line, we instanciate
# the main class and run it.
if __name__ == "__main__":
    SimpleExample().run()




#!/usr/bin/env python

"""OpenGL player for videos stored as directories of image files.

@author: Stephan Wenger
@date: 2012-02-29
"""

import os
from numpy import array
from PIL import Image

from glitter import Texture2D, get_copy_pipeline_2d
import glitter.raw as _gl
from glitter.contexts.glut import GlutWindow, main_loop

class ImageCache(object):
    def __init__(self, filenames):
        if not filenames:
            filenames = [os.getcwd()]

        def imopen(filename):
            try:
                return Image.open(filename)
            except IOError:
                return None

        def add_files(lst, dirname, fnames):
            lst.extend(x for x in (imopen(os.path.join(dirname, fname)) for fname in sorted(fnames)) if x)

        self.filenames = []
        for filename in filenames:
            if os.path.isfile(filename):
                self.filenames.append(Image.open(filename))
            elif os.path.isdir(filename):
                os.path.walk(filename, add_files, self.filenames)

    def __len__(self):
        return len(self.filenames)

    def __getitem__(self, idx):
        return array(self.filenames[idx]) # TODO caching and preloading

class IntroductionExample(object):
    fps_scale = 1.1

    def __init__(self, images):
        self.window = GlutWindow(double=True, multisample=True)
        self.window.display_callback = self.display
        self.window.keyboard_callback = self.keyboard
        self.window.special_callback = self.special
        self.images = images
        self.window.shape = self.images[0].shape[:2]
        self.render_pipeline = get_copy_pipeline_2d(context=self.window, sampler_type="isampler2D",
                maxval=255, yscale=-1, use_framebuffer=False, image=Texture2D(self._prepare_image(0)))
        self.current_index = 0
        self.fps = 25
        self.playing = False

    def _prepare_image(self, idx):
        img = self.images[idx]
        while img.ndim < 3:
            img = img[..., None]
        return img

    def display(self):
        self.window.clear()
        self.render_pipeline.image.data = self._prepare_image(self.current_index)
        self.window.window_title = self.images.filenames[self.current_index].filename
        self.render_pipeline.draw()
        self.window.swap_buffers()

    def keyboard(self, key, x, y):
        key = chr(key)
        if key == "+":
            self.fps *= self.fps_scale
        elif key == "-":
            self.fps /= self.fps_scale
        elif key == " ":
            if self.playing:
                self.playing = False
            else:
                self.playing = True
                self.timer()
        elif key == chr(27):
            raise SystemExit

    def special(self, key, x, y):
        if key == _gl.GLUT_KEY_LEFT:
            self.current_index = (self.current_index - 1) % len(self.images)
            self.window.post_redisplay()
        elif key == _gl.GLUT_KEY_RIGHT:
            self.current_index = (self.current_index + 1) % len(self.images)
            self.window.post_redisplay()
        if key == _gl.GLUT_KEY_PAGE_UP:
            self.current_index = (self.current_index - self.fps) % len(self.images)
            self.window.post_redisplay()
        elif key == _gl.GLUT_KEY_PAGE_DOWN:
            self.current_index = (self.current_index + self.fps) % len(self.images)
            self.window.post_redisplay()
        elif key == _gl.GLUT_KEY_HOME:
            self.current_index = 0
            self.window.post_redisplay()
        elif key == _gl.GLUT_KEY_END:
            self.current_index = len(self.images) - 1
            self.window.post_redisplay()

    def timer(self):
        if self.playing:
            self.current_index = (self.current_index + 1) % len(self.images)
            self.window.add_timer(int(round(1000.0 / self.fps)), self.timer)
            self.window.post_redisplay()

    def run(self):
        self.timer()
        main_loop()

if __name__ == "__main__":
    import sys
    IntroductionExample(ImageCache(sys.argv[1:])).run()




#!/usr/bin/env python

"""Mesh viewer that displays the depth instead of color.

@todo: Add literate programming comments.

@author: Stephan Wenger
@date: 2012-03-20
"""

from math import sin, cos, pi
from numpy.random import random
import h5py

from glitter import VertexArray, State, get_default_program, Framebuffer, Texture2D, get_copy_pipeline_2d
from glitter.contexts.glut import GlutWindow, main_loop, get_elapsed_time

class DepthViewer(object):
    def __init__(self, filename):
        self.window = GlutWindow(double=True, multisample=True)
        self.window.display_callback = self.display
        self.shader = get_default_program()
        self.fbo = Framebuffer(depth=Texture2D(shape=self.window.shape + (1,), depth=True))
        self.copy_pipeline = get_copy_pipeline_2d(image=self.fbo.depth, use_framebuffer=False)
        with h5py.File(filename, "r") as f:
            vertices = f["vertices"]
            colors = f.get("colors", None)
            elements = f.get("indices", None)
            if colors is None:
                colors = random((len(vertices), 3))[:, None, :][:, [0] * vertices.shape[1], :]
            self.vao = VertexArray(vertices, colors, elements=elements)

    def display(self):
        try:
            with self.fbo:
                self.fbo.clear()
                self.vao.draw()
            self.window.clear()
            self.copy_pipeline.draw()
            self.window.swap_buffers()
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise SystemExit(e)

    def timer(self):
        phi = 2 * pi * get_elapsed_time() / 20.0
        self.shader.modelview_matrix = ((cos(phi), 0, sin(phi), 0), (0, 1, 0, 0), (-sin(phi), 0, cos(phi), 0), (0, 0, 0, 1))
        self.window.add_timer(10, self.timer)
        self.window.post_redisplay()
    
    def run(self):
        self.timer()
        with self.shader:
            with State(depth_test=True):
                main_loop()

if __name__ == "__main__":
    import sys
    DepthViewer(sys.argv[1]).run()




"""Named constants.
"""

import re
import types

class NamedConstant(object):
    """Named constant.
    """

    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return "%s(%s)" % (self._name, super(NamedConstant, self).__repr__())

    def __str__(self):
        return self._name

class IntConstant(NamedConstant, int):
    """Named integer constant.
    """

    def __new__(cls, name, value):
        return int.__new__(cls, value)

    def __init__(self, name, value):
        NamedConstant.__init__(self, name)

    def __hash__(self):
        return int.__hash__(self)

    def _split_name(self):
        i = 0
        while i < len(self._name) and self._name[-i-1:].isdigit():
            i += 1
        basename = self._name[:-i]
        digit = int(self._name[-i:]) if i > 0 else None
        return basename, digit

    def __add__(self, other):
        if isinstance(other, NamedConstant):
            raise TypeError("cannot add two constants")
        else:
            basename, digit = self._split_name()
            if digit is None:
                raise TypeError("not a numeric constant")
            digit += other
            if digit < 0:
                raise ValueError("range exceeded")
            return type(self)("%s%d" % (basename, digit), super(NamedConstant, self).__add__(other))

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, NamedConstant):
            return super(NamedConstant, self).__sub__(other)
        else:
            basename, digit = self._split_name()
            if digit is None:
                raise TypeError("not a numeric constant")
            digit -= other
            if digit < 0:
                raise ValueError("range exceeded")
            return type(self)("%s%d" % (basename, digit), super(NamedConstant, self).__add__(other))

    def __or__(self, other):
        return type(self)("%s|%s" % (self, other), super(NamedConstant, self).__or__(other))

    def __ror__(self, other):
        return self | other

def make_constant(name, value):
    """Create a named constant with name C{name} and value C{value}.

    An appropriate wrapper class will be generated if necessary.
    """

    dtype = type(value)

    cls_name = "%sConstant" % "".join(x[0].upper() + x[1:] for x in dtype.__name__.split())
    if cls_name not in globals():
        d = {
            "__new__": lambda cls, name, value: dtype.__new__(cls, value),
            "__init__": lambda self, name, value: NamedConstant.__init__(self, name),
            "__hash__": lambda self: dtype.__hash__(self),
        }
        globals()[cls_name] = type(cls_name, (NamedConstant, dtype), d)

    return globals()[cls_name](name, value)

def wrap_constants(name_re="^(GL|GLU|GLUT|GLX)_[A-Z][A-Z0-9_]*$", types=(int, int, float), d=None):
    """Convert OpenGL constants to named constants.

    All values in C{d} that match C{name_re} and are of one of the types in
    C{types} will be replaced by corresponding L{NamedConstant}s. By default,
    C{d} is C{glitter.raw.__dict__}.
    """

    if d is None:
        from glitter import raw
        d = raw.__dict__

    for key, value in list(d.items()):
        if re.match(name_re, key) and isinstance(value, tuple(types)):
            d[key] = make_constant(key, value)

__all__ = ["NamedConstant", "wrap_constants"]




"""Array buffer class.

@todo: Implement slicing to allow for C{glVertexAttribPointer} with C{size}, C{stride}, and C{pointer} parameters.

@author: Stephan Wenger
@date: 2012-02-29
"""

import numpy as _np

import glitter.raw as _gl
from glitter.utils import buffer_dimensions_to_primitive, primitive_to_buffer_dimensions
from glitter.arrays.basebuffer import BaseBuffer

class ArrayBuffer(BaseBuffer):
    _binding = "array_buffer_binding"
    _target = _gl.GL_ARRAY_BUFFER

    def _use(self, index, num_components=None, stride=0, first=0):
        if num_components is None:
            if len(self.shape) == 1:
                num_components = 1
            elif 1 <= self.shape[-1] <= 4:
                num_components = self.shape[-1]
            else:
                raise ValueError("must specify num_components")
        if self.dtype.is_float():
            with self:
                _gl.glVertexAttribPointer(index, num_components, self.dtype._as_gl(), _gl.GL_FALSE, stride * self.dtype.nbytes, first * self.dtype.nbytes)
        else:
            with self:
                _gl.glVertexAttribIPointer(index, num_components, self.dtype._as_gl(), stride * self.dtype.nbytes, first * self.dtype.nbytes)

    def draw(self, mode=None, count=None, first=0, instances=None):
        if mode is None:
            if len(self.shape) > 2:
                mode = buffer_dimensions_to_primitive.get(self.shape[1], None)
            else:
                mode = buffer_dimensions_to_primitive.get(1, None)
        if mode is None:
            raise ValueError("must specify mode")

        if count is None:
            dim = primitive_to_buffer_dimensions.get(mode, None)
            if dim is not None and len(self.shape) > 2 and self.shape[-2] != dim:
                raise ValueError("buffer shape does not match mode")
            count = _np.prod(self.shape[:-1])
        
        if instances is None:
            with self:
                _gl.glDrawArrays(mode._value, first, count)
        else:
            with self:
                _gl.glDrawArraysInstances(mode._value, first, count, instances)

__all__ = ["ArrayBuffer"]




import numpy

from glitter import ArrayBuffer, Datatype, uint8, int8, uint16, int16, uint32, int32, float32

def check_buffer(shape, dtype, vrange):
    minval, maxval = vrange
    data = ((maxval - minval) * numpy.random.random(shape) + minval).astype(dtype.as_numpy())
    buf = ArrayBuffer(data)
    assert (buf.data == data).all(), "data is broken"
    assert buf.shape == data.shape, "shape is broken"
    assert buf.dtype == Datatype.from_numpy(data.dtype), "dtype is broken"
    assert buf._size == data.nbytes, "_size is broken"

def test_generator():
    shapes = ((4, 4, 4, 4), (4, 4, 4, 3), (4, 16, 8, 3), (5, 4, 4, 3), (5, 5, 5, 3), (6, 6, 6, 3), (7, 13, 5, 3), (1, 1, 3, 3))
    dtypes = (uint8, int8, uint16, int16, uint32, int32, float32)
    vranges = ((0, (1<<8)-1), (-1<<7, (1<<7)-1), (0, (1<<16)-1), (-1<<15, (1<<15)-1), (0, (1<<32)-1), (-1<<31, (1<<31)-1), (-10.0, 10.0))

    for shape in shapes:
        for dtype, vrange in zip(dtypes, vranges):
            yield check_buffer, shape, dtype, vrange




"""Automatic error checking for GL calls.
"""

import re

class GLError(Exception):
    """Exception representing an error within OpenGL.
    """

    def __init__(self, error, result, func, arguments):
        self.error = error
        self.result = result
        self.func = func
        self.arguments = arguments

    @staticmethod
    def string_from_id(error):
        """Lookup the error message for the given error number.
        """

        try:
            from glitter.raw import glu
            return glu.string_at(glu.gluErrorString(error))
        except (ImportError, AttributeError):
            return "error #%d" % error

    def __str__(self):
        return "GLError: %s in %s(%s)" % (GLError.string_from_id(self.error), self.func.__name__, ", ".join(repr(x) for x in self.arguments))

    def __repr__(self):
        return str(self)

def set_error_check(errcheck_func=Ellipsis, errcheck_ok=Ellipsis, name_re="^gl[A-Z].*$", d=None):
    """Add error handlers to OpenGL functions.

    The C{errcheck} attribute of all items in a dictionary C{d} having such an
    attribute and matching C{name_re} will be set to a function that calls
    C{errcheck_func} and raises L{GLError} if the return value is unequal to
    C{errcheck_ok}. By default, C{d} is C{glitter.raw.__dict__}.

    Defaults for C{errcheck_func} and C{errcheck_ok} are C{glGetError} and
    C{GL_NO_ERROR}, respectively. If C{errcheck_func} is C{None}, error
    checking will be disabled.
    """

    if d is None:
        from glitter import raw
        d = raw.__dict__

    if errcheck_func is Ellipsis:
        from glitter.raw import gl
        errcheck_func = gl.glGetError

    if errcheck_func is not None and errcheck_ok is Ellipsis:
        from glitter.raw import gl
        errcheck_ok = gl.GL_NO_ERROR

    def errcheck(result, func, arguments):
        error = errcheck_func()
        if error == errcheck_ok:
            return result
        else:
            raise GLError(error, result, func, arguments)

    for key, value in list(d.items()):
        if re.match(name_re, key) and value is not errcheck_func and hasattr(value, "errcheck"):
            value.errcheck = errcheck

__all__ = ["GLError", "set_error_check"]




from scipy.misc import imsave

from glitter import ShaderProgram, RectangleTexture, Framebuffer, VertexArray
from glitter.contexts.glut import GlutWindow, main_loop

vertex_shader = """
#version 400 core

layout(location=0) in vec4 in_position;

void main() {
    gl_Position = in_position;
}
"""

fragment_shader = """
#version 400 core

layout(location=0) out vec4 out_color;
uniform float dimx, dimy;

void main() {
    out_color = vec4(gl_FragCoord.x / dimx, gl_FragCoord.y / dimy, 1.0, 1.0);
}
"""

class MinimalFramebufferExample(object):
    def __init__(self):
        self.window = GlutWindow(double=True, multisample=True, shape=(100, 800))
        self.window.display_callback = self.display

        self.shader = ShaderProgram(vertex=vertex_shader, fragment=fragment_shader)
        self.shader.dimy, self.shader.dimx = self.window.shape
        self.fbo = Framebuffer(RectangleTexture(shape=self.window.shape + (3,)))

        self.vao = VertexArray(((-1.0, -1.0), (-1.0, 1.0), (1.0, 1.0), (1.0, -1.0)), elements=((0, 1, 2), (0, 2, 3)))

    def save(self, filename):
        self.fbo.clear()
        with self.shader:
            with self.fbo:
                self.vao.draw()
        imsave(filename, self.fbo[0].data)

    def display(self):
        self.window.clear()
        with self.shader:
            self.vao.draw()
        self.window.swap_buffers()

    def run(self):
        main_loop()

if __name__ == "__main__":
    ie = MinimalFramebufferExample()
    ie.save("test.png")
    ie.run()
    # TODO why do I get a bus error on exit?




"""OpenGL buffer objects.

@author: Stephan Wenger
@date: 2012-02-29
"""

from glitter.arrays.basebuffer import *
from glitter.arrays.arraybuffer import *
from glitter.arrays.elementarray import *
from glitter.arrays.vertexarray import *
from glitter.arrays.misc import *




"""Base classes for enumerations.

@author: Stephan Wenger
@date: 2012-02-29
"""

class EnumConstant(object):
    def __init__(self, enum, name, value):
        self._enum = enum
        self.name = name
        self._value = value

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

class Enum(object):
    def __init__(self, **kwargs):
        self._reverse_dict = {}
        for key, value in list(kwargs.items()):
            self._add(key, value)

    def __getitem__(self, value):
        return self._reverse_dict[value]

    def __call__(self, obj):
        """Convert C{obj} into an appropriate L{EnumConstant}.
        """

        if obj in list(self.__dict__.values()): # A matching EnumConstant is okay.
            return obj
        elif obj in list(self.__dict__.keys()): # A string is converted to an EnumConstant if possible.
            return self.__dict__[obj]
        else: # Anything else is just wrong.
            raise TypeError("'%s' is not a valid enum constant here" % obj)

    def _add(self, key, value):
        setattr(self, key, EnumConstant(self, key, value))
        self._reverse_dict[value] = getattr(self, key)

    def __repr__(self):
        return "Enum(%s)" % ", ".join(map(str, list(self._reverse_dict.values())))




import os
import re

from setuptools import Command
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as _build_py
from distutils import log

class generate_py_base(Command):
    user_options = [
        ("header=", None, "header file"),
        ("defines=", None, "C preprocessor variables to define"),
        ("include-dirs=", None, "C include directories (separated by %s)" % os.pathsep),
        ("libs=", None, "names of libraries to link against (separated by %s)" % os.pathsep),
        ("patterns=", None, "regular expressions for symbols to include (separated by whitespace)"),
        ("force", None, "rebuild even if target files exist"),
    ]
    
    def generate_xml(self, header, xml, defines=[], include_dirs=[]):
        commandline = ["h2xml.py", "-q", "-c", "-o", xml, header] + ["-D%s" % x for x in defines] + ["-I%s" % x for x in include_dirs]
        log.debug(" ".join(commandline))
        from ctypeslib import h2xml
        h2xml.compile_to_xml(commandline)

    def generate_python(self, xml, pythonname, libs=[], patterns=[]):
        commandline = ["xml2py.py", "-kamdefst", "-o", pythonname, xml] + ["-l%s" % x for x in libs] + ["-r%s" % x for x in patterns]
        log.debug(" ".join(commandline))
        from ctypeslib import xml2py
        xml2py.main(commandline)

    def initialize_options(self):
        self.header = None
        self.defines = []
        self.include_dirs = []
        self.libs = []
        self.patterns = []
        self.force = None

    def finalize_options(self):
        self.set_undefined_options("generate_py", ("force", "force"))
        if isinstance(self.defines, basestring):
            self.defines = re.findall("[A-Za-z0-9_]+", self.defines)
        if isinstance(self.include_dirs, basestring):
            self.include_dirs = self.include_dirs.split(os.pathsep)
        if isinstance(self.libs, basestring):
            self.libs = self.libs.split(os.pathsep)
        if isinstance(self.patterns, basestring):
            self.patterns = self.patterns.split()

    def run(self):
        basename = os.path.join("glitter", "raw", self.name)
        if self.distribution.package_dir is not None:
            basename = os.path.join(self.distutils.package_dir, basename)
        xmlname = basename + os.path.extsep + "xml"
        pyname = basename + os.path.extsep + "py"
        if self.force or not os.path.isfile(pyname):
            self.generate_xml(self.header, xmlname, self.defines, self.include_dirs)
            self.generate_python(xmlname, pyname, self.libs, self.patterns)
            os.remove(xmlname)

class generate_gl(generate_py_base):
    description = "generate ctypes wrapper for OpenGL"
    name = "gl"

    def initialize_options(self):
        generate_py_base.initialize_options(self)
        self.header = os.path.join(os.path.dirname(os.path.join(os.getcwd(), __file__)), "gl3.h")
        self.defines = ["GL3_PROTOTYPES"]
        self.include_dirs = ["/usr/include/GL"]
        self.libs = ["GL"]
        self.patterns = ['gl[A-Z].*', 'GL_[A-Z].*', 'GL[a-z].*']

class generate_glu(generate_py_base):
    description = "generate ctypes wrapper for GLU"
    name = "glu"

    def initialize_options(self):
        generate_py_base.initialize_options(self)
        self.header = "glu.h"
        self.include_dirs = ["/usr/include/GL"]
        self.libs = ["GLU"]
        self.patterns = ['glu[A-Z].*', 'GLU_[A-Z].*', 'GLU[a-z].*']
        
class generate_glut(generate_py_base):
    description = "generate ctypes wrapper for GLUT"
    name = "glut"

    def initialize_options(self):
        generate_py_base.initialize_options(self)
        self.header = "freeglut.h"
        self.include_dirs = ["/usr/include/GL"]
        self.libs = ["glut"]
        self.patterns = ['glut[A-Z].*', 'GLUT_[A-Z].*', 'GLUT[a-z].*']

class generate_glx(generate_py_base):
    description = "generate ctypes wrapper for GLX"
    name = "glx"

    def initialize_options(self):
        generate_py_base.initialize_options(self)
        self.header = "glx.h"
        self.include_dirs = ["/usr/include/GL"]
        self.libs = ["GL", "X11"]
        self.patterns = ['glX[A-Z].*', 'GLX_[A-Z].*', 'GLX[a-z].*']

class generate_py(Command):
    description = "generate ctypes wrappers"
    
    user_options = [
        ("force", None, "rebuild even if target files exist"),
    ]
    
    commands = ["generate_gl", "generate_glu", "generate_glut", "generate_glx"]

    def initialize_options(self):
        self.force = None

    def finalize_options(self):
        self.set_undefined_options("build", ("force", "force"))

    def run(self):
        for command in self.commands:
            self.run_command(command)

class build_py(_build_py):
    def run(self):
        self.run_command("generate_py")
        _build_py.run(self)

setup(
    name = "glitter",
    version = "0.1.6", # if you increase the version number, please update the change log in README.txt and add a git tag
    author = "Stephan Wenger",
    author_email = "wenger@cg.cs.tu-bs.de",
    description = "Intuitive OpenGL wrappers",
    license = "MIT",
    keywords = "opengl,graphics",
    long_description = open("README.txt").read(),
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 3 - Alpha",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Multimedia :: Graphics :: 3D Rendering",
        "Topic :: Software Development :: Libraries",
        ],
    download_url = "http://pypi.python.org/pypi/glitter",
    url = "http://packages.python.org/glitter",
    setup_requires = ["ctypeslib", "nose"],
    install_requires = ["numpy"],
    extras_require = {
        "Qt": ["PySide"],
        "HDF5": ["h5py"],
    },
    packages = find_packages(),
    test_suite = "nose.collector",
    cmdclass= {
        "generate_gl": generate_gl,
        "generate_glu": generate_glu,
        "generate_glut": generate_glut,
        "generate_glx": generate_glx,
        "generate_py": generate_py,
        "build_py": build_py,
    },
)




"""Miscellaneous buffer objects.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.arrays.basebuffer import BaseBuffer

class AtomicCounterBuffer(BaseBuffer):
    _binding = "atomic_counter_buffer_binding"
    _target = _gl.GL_ATOMIC_COUNTER_BUFFER

class CopyReadBuffer(BaseBuffer):
    _binding = "copy_read_buffer_binding"
    _target = _gl.GL_COPY_READ_BUFFER

class CopyWriteBuffer(BaseBuffer):
    _binding = "copy_write_buffer_binding"
    _target = _gl.GL_COPY_WRITE_BUFFER

class DrawIndirectBuffer(BaseBuffer):
    _binding = "draw_indirect_buffer_binding"
    _target = _gl.GL_DRAW_INDIRECT_BUFFER

class PixelPackBuffer(BaseBuffer):
    _binding = "pixel_pack_buffer_binding"
    _target = _gl.GL_PIXEL_PACK_BUFFER

class PixelUnpackBuffer(BaseBuffer):
    _binding = "pixel_unpack_buffer_binding"
    _target = _gl.GL_PIXEL_UNPACK_BUFFER

class TextureBuffer(BaseBuffer):
    _binding = "texture_buffer_binding"
    _target = _gl.GL_TEXTURE_BUFFER

class TransformFeedbackBuffer(BaseBuffer):
    _binding = "transform_feedback_buffer_binding"
    _target = _gl.GL_TRANSFORM_FEEDBACK_BUFFER

class UniformBuffer(BaseBuffer):
    _binding = "uniform_buffer_binding"
    _target = _gl.GL_UNIFORM_BUFFER

__all__ = [
    "AtomicCounterBuffer",
    "CopyReadBuffer",
    "CopyWriteBuffer",
    "DrawIndirectBuffer",
    "PixelPackBuffer",
    "PixelUnpackBuffer",
    "TextureBuffer",
    "TransformFeedbackBuffer",
    "UniformBuffer",
]




"""Query classes.

@warning: The C{Query} class is currently untested.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, BindReleaseObject

class Query(ManagedObject, BindReleaseObject):
    _generate_id = _gl.glGenQueries
    _delete_id = _gl.glDeleteQueries
    _db = "queries"

    def __init__(self, context=None):
        if any(x is NotImplemented for x in (self._target,)):
            raise TypeError("%s is abstract" % self.__class__.__name__)
        super(Query, self).__init__(context=context)
        if self._counter_bits == 0:
            raise RuntimeError("%s not supported" % self.__class__.__name__)

    def bind(self):
        _gl.glBeginQuery(self._id)

    def release(self):
        _gl.glEndQuery()

    @property
    def _current_query(self):
        _current_query = _gl.GLint()
        _gl.glGetQueryiv(self._target, _gl.GL_CURRENT_QUERY, _gl.pointer(_current_query))
        return _current_query.value

    @property
    def _counter_bits(self):
        _counter_bits = _gl.GLint()
        _gl.glGetQueryiv(self._target, _gl.GL_QUERY_COUNTER_BITS, _gl.pointer(_counter_bits))
        return _counter_bits.value

    @property
    def result_available(self):
        _result_available = _gl.GLint()
        _gl.glGetQueryObjectiv(self._id, _gl.GL_QUERY_RESULT_AVAILABLE, _gl.pointer(_result_available))
        return bool(_result_available.value)

    @property
    def result(self):
        if self._counter_bits <= 32:
            _result = _gl.GLuint()
            _gl.glGetQueryObjectuiv(self._id, _gl.GL_QUERY_RESULT_AVAILABLE, _gl.pointer(_result))
        else:
            _result = _gl.GLuint64()
            _gl.glGetQueryObjectui64v(self._id, _gl.GL_QUERY_RESULT_AVAILABLE, _gl.pointer(_result))
        return _result.value

class SamplesPassedQuery(Query):
    """@todo: Wrap C{glBeginConditionalRender} and C{glEndConditionalRender}."""
    _target = _gl.GL_SAMPLES_PASSED

class AnySamplesPassedQuery(Query):
    """@todo: Wrap C{glBeginConditionalRender} and C{glEndConditionalRender}."""
    _target = _gl.GL_ANY_SAMPLES_PASSED
    result = property(lambda self: bool(Query.result.fget(self)))

class PrimitivesGeneratedQuery(Query):
    _target = _gl.GL_PRIMITIVES_GENERATED

class TransformFeedbackPrimitivesWrittenQuery(Query):
    _target = _gl.GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN

class TimeElapsedQuery(Query):
    _target = _gl.GL_TIME_ELAPSED

__all__ = [
    "Query",
    "SamplesPassedQuery",
    "AnySamplesPassedQuery",
    "PrimitivesGeneratedQuery",
    "TransformFeedbackPrimitivesWrittenQuery",
    "TimeElapsedQuery",
]




"""Basic ctypes wrappers around OpenGL 3 (and up).

This module provides very basic wrapping of OpenGL, GLU, GLUT, and GLX. Like
PyOpenGL, it includes typesafe function calls, named constants and automatic
error checking. Since it is mainly intended as a low-level backend for the
glitter package, it does not provide wrapping of return types like PyOpenGL,
but instead includes logging facilities for easier OpenGL debugging through a
high-level API.
"""

from ctypes import *

try:
    from glitter.raw.gl import *
except ImportError:
    pass

try:
    from glitter.raw.glu import *
except ImportError:
    pass

try:
    from glitter.raw.glut import *
except ImportError:
    pass

try:
    from glitter.raw.glx import *
except ImportError:
    pass

from glitter.raw.errcheck import *
from glitter.raw.logger import *
from glitter.raw.constant import *

wrap_constants(d=globals()) # pass d explicitly to avoid infinite import loop
set_error_check(d=globals()) # pass d explicitly to avoid infinite import loop

